# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore import BaseAnalysis, NetworkDataset
from pyincore.utils.networkutil import NetworkUtil
from numpy.linalg import inv
from typing import List
from scipy import stats
import networkx as nx
import pandas as pd
import numpy as np
import copy


class NciFunctionality(BaseAnalysis):
    """
    This analysis computes the output of the Leontief equation for functional dependencies between two
    interdependent networks having functionality information per node. These dependencies capture cascading
    dependencies on infrastructure functionality, expressed in terms of discrete points.

    The output of the computation consists of two datasets, one per each labeled network, with new cascading
    functionalities accompanying the original discrete ones.

    Contributors
        | Science: Milad Roohi, John van de Lindt
        | Implementation: Milad Roohi, Santiago Núñez-Corrales and NCSA IN-CORE Dev Team

    Related publications

        Roohi M, van de Lindt JW, Rosenheim N, Hu Y, Cutler H. (2021) Implication of building inventory accuracy
        on physical and socio-economic resilience metrics for informed decision-making in natural hazards.
        Structure and Infrastructure Engineering. 2020 Nov 20;17(4):534-54.

        Milad Roohi, Jiate Li, John van de Lindt. (2022) Seismic Functionality Analysis of Interdependent
        Buildings and Lifeline Systems 12th National Conference on Earthquake Engineering (12NCEE),
        Salt Lake City, UT (June 27-July 1, 2022).


    Args:

        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(NciFunctionality, self).__init__(incore_client)

    def run(self):
        # Load parameters
        discretized_days = self.get_parameter('discretized_days')

        # Load all dataset-related entities for EPF
        epf_network_dataset = NetworkDataset.from_dataset(self.get_input_dataset('epf_network'))
        epf_network_nodes = epf_network_dataset.nodes.get_dataframe_from_shapefile()
        epf_network_links = epf_network_dataset.links.get_dataframe_from_shapefile()

        # Load all dataset-related entities for WDS
        wds_network_dataset = NetworkDataset.from_dataset(self.get_input_dataset('wds_network'))
        wds_network_nodes = wds_network_dataset.nodes.get_dataframe_from_shapefile()
        wds_network_links = wds_network_dataset.links.get_dataframe_from_shapefile()

        # Load network interdependencies
        epf_wds_intdp_table = self.get_input_dataset('epf_wds_intdp_table').get_dataframe_from_csv()
        wds_epf_intdp_table = self.get_input_dataset('wds_epf_intdp_table').get_dataframe_from_csv()

        # Load restoration functionality and time results for EPF
        epf_time_results = self.get_input_dataset('epf_time_results').get_dataframe_from_csv()
        epf_subst_failure_results = self.get_input_dataset('epf_subst_failure_results').get_dataframe_from_csv()
        epf_inventory_rest_map = self.get_input_dataset('epf_inventory_rest_map').get_dataframe_from_csv()

        # Load restoration functionality and time results for WDS
        wds_time_results = self.get_input_dataset('wds_time_results').get_dataframe_from_csv()
        wds_dmg_results = self.get_input_dataset('wds_dmg_results').get_dataframe_from_csv()
        wds_inventory_rest_map = self.get_input_dataset('wds_inventory_rest_map').get_dataframe_from_csv()

        epf_cascading_functionality = self.nci_functionality(discretized_days, epf_network_nodes, epf_network_links,
                                                             wds_network_nodes, wds_network_links,
                                                             epf_wds_intdp_table, wds_epf_intdp_table,
                                                             epf_subst_failure_results, epf_inventory_rest_map,
                                                             epf_time_results, wds_dmg_results,  wds_inventory_rest_map,
                                                             wds_time_results)

        result_name = self.get_parameter("result_name")
        self.set_result_csv_data("epf_cascading_functionality", epf_cascading_functionality, name=result_name,
                                 source="dataframe")

        return True

    def nci_functionality(self, discretized_days, epf_network_nodes, epf_network_links, wds_network_nodes,
                          wds_network_links, epf_wds_intdp_table, wds_epf_intdp_table, epf_subst_failure_results,
                          epf_inventory_rest_map, epf_time_results, wds_dmg_results, wds_inventory_rest_map,
                          wds_time_results):
        """Compute EPF and WDS cascading functionality outcomes

        Args:
            discretized_days (List[int]): a list of discretized days
            epf_network_nodes (pd.DataFrame): network nodes for EPF network
            epf_network_links (pd.DataFrame): network links for EPF network
            wds_network_nodes (pd.DataFrame): network nodes for WDS network
            wds_network_links (pd.DataFrame): network links for WDS network
            epf_wds_intdp_table (pd.DataFrame): mapping from EPF to WDS networks
            wds_epf_intdp_table (pd.DataFrame): mapping from WDS to EPF networks
            epf_subst_failure_results (pd.DataFrame): substation failure results for EPF network
            epf_inventory_rest_map (pd.DataFrame): inventory restoration map for EPF network
            epf_time_results (pd.DataFrame): time results for EPF network
            wds_dmg_results (pd.DataFrame): damage results for WDS network
            wds_inventory_rest_map (pd.DataFrame): inventory restoration map for WDS network
            wds_time_results (pd.DataFrame): time results for WDS network

        Returns:
            (pd.DataFrame, pd.DataFrame): results for EPF and WDS networks
        """

        # Compute updated EPF and WDS node information
        efp_nodes_updated = self.update_epf_discretized_func(epf_network_nodes, epf_subst_failure_results,
                                                             epf_inventory_rest_map, epf_time_results)

        wds_nodes_updated = self.update_wds_discretized_func(wds_network_nodes, wds_dmg_results,
                                                             wds_inventory_rest_map, wds_time_results)

        # Compute updated WDS links
        wds_links_updated = self.update_wds_network_links(wds_network_links)

        # Generate the functionality data
        df_functionality_nodes = pd.concat([efp_nodes_updated, wds_nodes_updated], ignore_index=True)
        df_functionality_nodes = pd.concat([efp_nodes_updated, wds_nodes_updated], ignore_index=True)

        # Create each individual graph
        g_epf = NetworkUtil.create_network_graph_from_dataframes(epf_network_nodes, epf_network_links)
        g_wds = NetworkUtil.create_network_graph_from_dataframes(wds_network_nodes, wds_links_updated)

        # Obtain two graphs for directional interdependencies
        g_epf_wds = NetworkUtil.merge_labeled_networks(g_epf, g_wds, epf_wds_intdp_table, directed=True)

        # To be implemented in a future release
        # g_wds_epf = NetworkUtil.merge_labeled_networks(g_wds, g_epf, wds_epf_intdp_table, directed=True)

        # Solve the corresponding Leontief problems
        df_epn_func_nodes = self.solve_leontief_equation(g_epf_wds, df_functionality_nodes, discretized_days)

        # To be implemented in a future release
        # df_wds_func_nodes = self.solve_leontief_equation(g_wds_epf, df_functionality_nodes, discretized_days)

        epn_cascading_functionality = epf_network_nodes[['guid', 'geometry']].merge(df_epn_func_nodes, on='guid',
                                                                                    how='left').rename(columns={
            'guid': 'sguid'})

        # To be implemented in a future release
        # wds_cascading_functionality = wds_network_nodes[['guid', 'geometry']].merge(df_wds_func_nodes, on='guid',
                                                                                    #how='left').rename(columns={
           #'guid': 'sguid'})

        return epn_cascading_functionality

    @staticmethod
    def update_epf_discretized_func(epf_nodes, epf_subst_failure_results, epf_inventory_restoration_map,
                                    epf_time_results):
        epf_time_results = epf_time_results.loc[
            (epf_time_results['time'] == 1) | (epf_time_results['time'] == 3) | (epf_time_results['time'] == 7) | (
                        epf_time_results['time'] == 30) | (epf_time_results['time'] == 90)]
        epf_time_results.insert(2, 'PF_00', list(
            np.ones(len(epf_time_results))))  # PF_00, PF_0, PF_1, PF_2, PF_3  ---> DS_0, DS_1, DS_2, DS_3, DS_4

        epf_nodes_updated = pd.merge(epf_nodes[['nodenwid', 'utilfcltyc', 'guid']], epf_subst_failure_results[
            ['guid', 'DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4', 'failure_probability']], on='guid', how='outer')

        EPPL_restoration_id = list(epf_inventory_restoration_map.loc[epf_inventory_restoration_map['guid'] ==
                                                                     epf_nodes_updated.loc[epf_nodes_updated[
                                                                                               'utilfcltyc'] == 'EPPL'].guid.tolist()[
                                                                         0]]['restoration_id'])[0]
        ESS_restoration_id = \
        list(set(epf_inventory_restoration_map.restoration_id.unique()) - set([EPPL_restoration_id]))[0]
        df_EPN_node_EPPL = epf_nodes_updated.loc[epf_nodes_updated['utilfcltyc'] == 'EPPL']
        df_EPN_node_ESS = epf_nodes_updated.loc[epf_nodes_updated['utilfcltyc'] != 'EPPL']
        epf_time_results_EPPL = epf_time_results.loc[epf_time_results['restoration_id'] == EPPL_restoration_id][
            ['PF_00', 'PF_0', 'PF_1', 'PF_2', 'PF_3']]
        EPPL_func_df = pd.DataFrame(
            np.dot(df_EPN_node_EPPL[['DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4']], np.array(epf_time_results_EPPL).T),
            columns=['functionality1', 'functionality3', 'functionality7', 'functionality30', 'functionality90'])
        EPPL_func_df.insert(0, 'guid', list(df_EPN_node_EPPL.guid))
        epf_time_results_ESS = epf_time_results.loc[epf_time_results['restoration_id'] == ESS_restoration_id][
            ['PF_00', 'PF_0', 'PF_1', 'PF_2', 'PF_3']]
        ESS_func_df = pd.DataFrame(
            np.dot(df_EPN_node_ESS[['DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4']], np.array(epf_time_results_ESS).T),
            columns=['functionality1', 'functionality3', 'functionality7', 'functionality30', 'functionality90'])
        ESS_func_df.insert(0, 'guid', list(df_EPN_node_ESS.guid))
        epf_function_df = pd.concat([ESS_func_df, EPPL_func_df], ignore_index=True)
        epf_nodes_updated = pd.merge(epf_nodes_updated, epf_function_df, on="guid")

        return epf_nodes_updated

    @staticmethod
    def update_wds_discretized_func(wds_nodes, wds_dmg_results, wds_inventory_restoration_map, wds_time_results):
        wf_time_results = wds_time_results.loc[
            (wds_time_results['time'] == 1) | (wds_time_results['time'] == 3) | (wds_time_results['time'] == 7) | (
                        wds_time_results['time'] == 30) | (wds_time_results['time'] == 90)]
        wf_time_results.insert(2, 'PF_00', list(np.ones(len(wf_time_results))))


        wds_nodes_updated = pd.merge(wds_nodes[['nodenwid', 'utilfcltyc', 'guid']],
                                        wds_dmg_results[['guid', 'DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4']], on='guid',
                                        how='outer')

        PPPL_restoration_id = list(wds_inventory_restoration_map.loc[wds_inventory_restoration_map['guid'] ==
                                                                    wds_nodes_updated.loc[wds_nodes_updated[
                                                                                        'utilfcltyc'] == 'PPPL'].guid.tolist()[
                                                                        0]]['restoration_id'])[0]
        PSTAS_restoration_id = list(wds_inventory_restoration_map.loc[wds_inventory_restoration_map['guid'] ==
                                                                     wds_nodes_updated.loc[wds_nodes_updated[
                                                                                         'utilfcltyc'] == 'PSTAS'].guid.tolist()[
                                                                         0]]['restoration_id'])[0]
        df_wds_node_PPPL = wds_nodes_updated.loc[wds_nodes_updated['utilfcltyc'] == 'PPPL']
        df_wds_node_PSTAS = wds_nodes_updated.loc[wds_nodes_updated['utilfcltyc'] == 'PSTAS']

        wf_time_results_PPPL = wf_time_results.loc[wf_time_results['restoration_id'] == PPPL_restoration_id][
            ['PF_00', 'PF_0', 'PF_1', 'PF_2', 'PF_3']]
        PPPL_func_df = pd.DataFrame(
            np.dot(df_wds_node_PPPL[['DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4']], np.array(wf_time_results_PPPL).T),
            columns=['functionality1', 'functionality3', 'functionality7', 'functionality30', 'functionality90'])
        PPPL_func_df.insert(0, 'guid', list(df_wds_node_PPPL.guid))

        wf_time_results_PSTAS = wf_time_results.loc[wf_time_results['restoration_id'] == PSTAS_restoration_id][
            ['PF_00', 'PF_0', 'PF_1', 'PF_2', 'PF_3']]
        PSTAS_func_df = pd.DataFrame(
            np.dot(df_wds_node_PSTAS[['DS_0', 'DS_1', 'DS_2', 'DS_3', 'DS_4']], np.array(wf_time_results_PSTAS).T),
            columns=['functionality1', 'functionality3', 'functionality7', 'functionality30', 'functionality90'])
        PSTAS_func_df.insert(0, 'guid', list(df_wds_node_PSTAS.guid))

        wf_function_df = pd.concat([PSTAS_func_df, PPPL_func_df], ignore_index=True)
        wds_nodes_updated = pd.merge(wds_nodes_updated, wf_function_df, on="guid")

        return wds_nodes_updated

    @staticmethod
    def solve_leontief_equation(graph, functionality_nodes, discretized_days):
        """Computes the solution to the Leontief equation for network interdependency given a

        Args:
            graph (networkx object): graph containing the integrated EPN-WDS network
            functionality_nodes (pd.DataFrame): dataframe containing discretized EFP/WDS restoration results
            per node
            discretized_days (list): days used for discretization of restoration analyses

        Returns:
            pd.DataFrame
        """
        # Create a deep copy of the incoming fuctionality results to store new values
        df_functionality_nodes = copy.deepcopy(functionality_nodes)

        # M is computed once and remains common across discretized days
        M = nx.adjacency_matrix(graph).todense()

        for idx in discretized_days:
            u = 1 - df_functionality_nodes[f'functionality{idx}']
            u = u.to_numpy()
            I = np.identity(len(u))
            q = np.dot(np.linalg.inv(I - M.T), u).tolist()
            df_functionality_nodes[f'func_cascading{idx}'] = [0 if i >= 1 else 1 - i for i in q]

        return df_functionality_nodes

    @staticmethod
    def update_wds_network_links(wds_network_links):
        """Update network links with functionality attributes

        Args:
            wds_network_links (pd.DataFrame): WDS network links

        Returns:
            pd.DataFrame

        """
        wds_links = copy.deepcopy(wds_network_links)

        # Use `numpgvrpr` from pipeline damage
        #wds_links = pd.merge(wds_links, pp_dmg_results, on='guid', how='outer')

        # Update values with pgv and pgd calculations
        for idx in wds_links['linknwid']:
            df = wds_links[wds_links.linknwid.isin([idx])]

            # standard deviation of normal distribution
            sigma = 0.85

            # mean of normal distribution
            mu = np.log(.1)

            C_pgv = 0.2  # 0.2
            C_pgd = 0.8  # 0.8
            im = (C_pgv * df['numpgvrpr'] + C_pgd * df['numpgdrpr']).sum() / df['length'].sum()
            SI_break = 1 - stats.lognorm(s=sigma, scale=np.exp(mu)).cdf(im)

            C_pgv = 0.8  # 0.2
            C_pgd = 0.2  # 0.8
            im = (C_pgv * df['numpgvrpr'] + C_pgd * df['numpgdrpr']).sum() / df['length'].sum()
            SI_leak = 1 - stats.lognorm(s=sigma, scale=np.exp(mu)).cdf(im)

            m = wds_links['linknwid'] == idx
            wds_links.loc[m, ['SI_break_idv']] = SI_break
            wds_links.loc[m, ['SI_leak__idv']] = SI_leak

            return wds_links

    def get_spec(self):
        """Get specifications of the network cascading interdependency functionality analysis.
        Returns:
            obj: A JSON object of specifications of the NCI functionality analysis.
        """
        return {
            'name': 'network-cascading-interdepedency-functionality',
            'description': 'Network cascading interdepedency functionality analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'discretized_days',
                    'required': False,
                    'description': 'Discretized days to compute functionality',
                    'type': List[int]
                }
            ],
            'input_datasets': [
                {
                    'id': 'epf_network',
                    'required': True,
                    'description': 'EPN network',
                    'type': ['incore:epnNetwork'],
                },
                {
                    'id': 'wds_network',
                    'required': True,
                    'description': 'WDS network',
                    'type': ['incore:waterNetwork'],
                },
                {
                    'id': 'epf_wds_intdp_table',
                    'required': True,
                    'description': 'Table containing interdependency information from EPN to WDS networks',
                    'type': ['incore:networkInterdependencyTable']
                },
                {
                    'id': 'wds_epf_intdp_table',
                    'required': True,
                    'description': 'Table containing interdependency information from WDS to EPF networks',
                    'type': ['incore:networkInterdependencyTable']
                },
                {
                    'id': 'epf_subst_failure_results',
                    'required': True,
                    'description': 'EPF substation failure results',
                    'type': ['incore:failureProbability']
                },
                {
                    'id': 'epf_inventory_rest_map',
                    'required': True,
                    'description': 'EPF inventory restoration map',
                    'type': ['incore:inventoryRestorationMap']
                },
                {

                    'id': 'epf_time_results',
                    'required': True,
                    'description': 'A csv file recording repair time for EPF per class and limit state',
                    'type': ['incore:epfRestorationTime']
                },
                {
                    'id': 'wds_dmg_results',
                    'required': True,
                    'description': 'WDS damage results',
                    'type': ['ergo:waterFacilityDamageVer6']
                },
                {
                    'id': 'wds_inventory_rest_map',
                    'required': True,
                    'description': 'WDS inventory restoration map',
                    'type': ['incore:inventoryRestorationMap']
                },
                {

                    'id': 'wds_time_results',
                    'required': True,
                    'description': 'A csv file recording repair time for WDS per class and limit state',
                    'type': ['incore:waterFacilityRestorationTime']
                }
            ],
            'output_datasets': [
                {
                    'id': 'epf_cascading_functionality',
                    'description': 'CSV file of interdependent cascading network functionality for EPF',
                    'type': 'incore:epfDiscretizedCascadingFunc'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


""" Buried Pipeline Damage Analysis with limit state calculation """

import concurrent.futures
from itertools import repeat

from pyincore import BaseAnalysis, HazardService, FragilityService, \
     AnalysisUtil, GeoUtil
from pyincore.models.dfr3curve import DFR3Curve


class PipelineDamage(BaseAnalysis):
    """Computes pipeline damage for an earthquake or a tsunami).

    Args:
        incore_client: Service client with authentication info.

    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(PipelineDamage, self).__init__(incore_client)

    def run(self):
        """Execute pipeline damage analysis """

        pipeline_dataset = self.get_input_dataset("pipeline").get_inventory_reader()

        # Get hazard input
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        dataset_size = len(pipeline_dataset)
        num_workers = AnalysisUtil.determine_parallelism_locally(self,
                                                                 dataset_size,
                                                                 user_defined_cpu)
        avg_bulk_input_size = int(dataset_size / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(pipeline_dataset)
        while count < len(inventory_list):
            inventory_args.append(
                inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (results, damage_results) = self.pipeline_damage_concurrent_future(
            self.pipeline_damage_analysis_bulk_input, num_workers,
            inventory_args, repeat(hazard), repeat(hazard_type), repeat(hazard_dataset_id))

        self.set_result_csv_data("result", results, name=self.get_parameter("result_name"))
        self.set_result_json_data("metadata",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")
        return True

    def pipeline_damage_concurrent_future(self, function_name, num_workers,
                                          *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            dict: An ordered dictionaries with pipeline damage values.
            dict: An ordered dictionaries with other pipeline data/metadata.

        """
        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=num_workers) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def pipeline_damage_analysis_bulk_input(self, pipelines, hazard, hazard_type, hazard_dataset_id):
        """Run pipeline damage analysis for multiple pipelines.

        Args:
            pipelines (list): Multiple pipelines from pipeline dataset.
            hazard (obj): Hazard object.
            hazard_type (str): Hazard type (earthquake or tsunami).
            hazard_dataset_id (str): An id of the hazard exposure.

        Returns:
            dict: An ordered dictionaries with pipeline damage values.
            dict: An ordered dictionaries with other pipeline data/metadata.

        """
        # get allowed demand types for the hazard type
        allowed_demand_types = [item["demand_type"].lower() for item in self.hazardsvc.get_allowed_demands(
            hazard_type)]

        # Get Fragility key
        fragility_key = self.get_parameter("fragility_key")
        if fragility_key is None:
            fragility_key = "Non-Retrofit inundationDepth Fragility ID Code" if hazard_type == 'tsunami' else "pgv"
            self.set_parameter("fragility_key", fragility_key)

        # get fragility set
        fragility_sets = self.fragilitysvc.match_inventory(
            self.get_input_dataset("dfr3_mapping_set"), pipelines, fragility_key)

        values_payload = []
        unmapped_pipelines = []
        mapped_pipelines = []
        for pipeline in pipelines:
            # if find a match fragility for that pipeline
            if pipeline["id"] in fragility_sets.keys():
                fragility_set = fragility_sets[pipeline["id"]]
                location = GeoUtil.get_location(pipeline)
                loc = str(location.y) + "," + str(location.x)
                demands, units, _ = AnalysisUtil.get_hazard_demand_types_units(pipeline, fragility_set, hazard_type,
                                                                               allowed_demand_types)
                value = {
                    "demands": demands,
                    "units": units,
                    "loc": loc
                }
                values_payload.append(value)
                mapped_pipelines.append(pipeline)

            else:
                unmapped_pipelines.append(pipeline)

        # not needed anymore as they are already split into mapped and unmapped
        del pipelines

        if hazard_type == 'earthquake' or "tsunami":
            hazard_vals = hazard.read_hazard_values(values_payload, self.hazardsvc)
        else:
            raise ValueError("The provided hazard type is not supported yet by this analysis")

        pipeline_results = []
        damage_results = []
        for i, pipeline in enumerate(mapped_pipelines):
            limit_states = dict()
            dmg_intervals = dict()
            pipeline_result = dict()
            fragility_set = fragility_sets[pipeline["id"]]

            # TODO: Once all fragilities are migrated to new format, we can remove this condition
            if isinstance(fragility_set.fragility_curves[0], DFR3Curve):
                # Supports multiple demand types in same fragility
                haz_vals = AnalysisUtil.update_precision_of_lists(hazard_vals[i]["hazardValues"])
                demand_types = hazard_vals[i]["demands"]
                demand_units = hazard_vals[i]["units"]

                # construct hazard_value dictionary {"demand_type":"hazard_value", ...}
                hval_dict = dict()
                for j, d in enumerate(fragility_set.demand_types):
                    hval_dict[d] = haz_vals[j]

                if not AnalysisUtil.do_hazard_values_have_errors(hazard_vals[i]["hazardValues"]):
                    pipeline_args = fragility_set.construct_expression_args_from_inventory(pipeline)
                    limit_states = fragility_set.calculate_limit_state(hval_dict,
                                                                       inventory_type="pipeline",
                                                                       **pipeline_args)
                    dmg_intervals = fragility_set.calculate_damage_interval(limit_states, hazard_type=hazard_type,
                                                                            inventory_type="pipeline")

            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            pipeline_result['guid'] = pipeline['properties']['guid']
            pipeline_result.update(limit_states)
            pipeline_result.update(dmg_intervals)
            pipeline_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(haz_vals, hazard_type)
            damage_result = dict()
            damage_result['guid'] = pipeline['properties']['guid']
            damage_result['fragility_id'] = fragility_set.id
            damage_result['demandtypes'] = demand_types
            damage_result['demandunits'] = demand_units
            damage_result['hazardtype'] = hazard_type
            damage_result['hazardval'] = haz_vals

            pipeline_results.append(pipeline_result)
            damage_results.append(damage_result)

        # for pipeline does not have matching fragility curves, default to None
        for pipeline in unmapped_pipelines:
            pipeline_result = dict()
            damage_result = dict()
            pipeline_result['guid'] = pipeline['properties']['guid']
            damage_result['guid'] = pipeline['properties']['guid']
            damage_result['fragility_id'] = None
            damage_result['demandtypes'] = None
            damage_result['demandunits'] = None
            damage_result['hazardtype'] = None
            damage_result['hazardvals'] = None

            pipeline_results.append(pipeline_result)
            damage_results.append(damage_result)

        return pipeline_results, damage_results

    def get_spec(self):
        """Get specifications of the pipeline damage analysis.

        Returns:
            obj: A JSON object of specifications of the pipeline damage analysis.

        """
        return {
            'name': 'pipeline-damage',
            'description': 'Buried pipeline damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result dataset name',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard Type',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Geology dataset id',
                    'type': str,
                }
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'pipeline',
                    'required': True,
                    'description': 'Pipeline Inventory',
                    'type': ['ergo:buriedPipelineTopology', 'ergo:pipeline'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'pipeline',
                    'description': 'CSV file of damage states for pipeline damage',
                    'type': 'incore:pipelineDamageVer3'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'pipeline',
                    'description': 'Json file with information about applied hazard value and fragility',
                    'type': 'incore:pipelineDamageSupplement'
                }
            ]
        }



from pyincore.analyses.buildingportfolio.BuildingPortfolioRecoveryAnalysis import BuildingPortfolioRecoveryAnalysis
from pyincore.analyses.buildingportfolio.BuildingDamage import BuildingDamage
from pyincore.analyses.buildingportfolio.BuildingData import BuildingData



class BuildingDamage(object):

    def __init__(self, distance_to_center, restricted_entry, restricted_use, reoccupancy, best_line_functionality,
                 full_functionality):
        self.distance_to_center = distance_to_center
        self.restricted_entry = restricted_entry
        self.restricted_use = restricted_use
        self.reoccupancy = reoccupancy
        self.best_line_functionality = best_line_functionality
        self.full_functionality = full_functionality



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import concurrent.futures
from itertools import repeat

from pyincore import AnalysisUtil, GeoUtil
from pyincore import BaseAnalysis, HazardService, FragilityService
from pyincore.analyses.bridgedamage.bridgeutil import BridgeUtil
from pyincore.models.dfr3curve import DFR3Curve


class BridgeDamage(BaseAnalysis):
    """Computes bridge structural damage for earthquake, tsunami, tornado, and hurricane hazards.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(BridgeDamage, self).__init__(incore_client)

    def run(self):
        """Executes bridge damage analysis."""
        # Bridge dataset
        bridge_set = self.get_input_dataset("bridges").get_inventory_reader()

        # get input hazard
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(
            bridge_set), user_defined_cpu)

        avg_bulk_input_size = int(len(bridge_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(bridge_set)
        while count < len(inventory_list):
            inventory_args.append(
                inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.bridge_damage_concurrent_future(
            self.bridge_damage_analysis_bulk_input, num_workers, inventory_args, repeat(hazard), repeat(hazard_type),
            repeat(hazard_dataset_id))

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("metadata",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")

        return True

    def bridge_damage_concurrent_future(self, function_name, num_workers,
                                        *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with bridge damage values and other data/metadata.

        """
        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def bridge_damage_analysis_bulk_input(self, bridges, hazard, hazard_type, hazard_dataset_id):
        """Run analysis for multiple bridges.

        Args:
            bridges (list): Multiple bridges from input inventory set.
            hazard (obj): Hazard object.
            hazard_type (str): Type of hazard.
            hazard_dataset_id (str): ID of hazard.

        Returns:
            list: A list of ordered dictionaries with bridge damage values and other data/metadata.

        """

        # Get Fragility key
        fragility_key = self.get_parameter("fragility_key")
        if fragility_key is None:
            fragility_key = BridgeUtil.DEFAULT_TSUNAMI_HMAX_FRAGILITY_KEY if hazard_type == 'tsunami' else \
                BridgeUtil.DEFAULT_FRAGILITY_KEY
            self.set_parameter("fragility_key", fragility_key)

        # Hazard Uncertainty
        use_hazard_uncertainty = False
        if hazard_type == "earthquake" and self.get_parameter(
                "use_hazard_uncertainty") is not None:
            use_hazard_uncertainty = self.get_parameter(
                "use_hazard_uncertainty")

        # Liquefaction
        use_liquefaction = False
        if hazard_type == "earthquake" and self.get_parameter(
                "use_liquefaction") is not None:
            use_liquefaction = self.get_parameter("use_liquefaction")

        # Get geology dataset id containing liquefaction susceptibility
        geology_dataset_id = self.get_parameter("liquefaction_geology_dataset_id")

        fragility_set = self.fragilitysvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), bridges,
                                                          fragility_key)

        values_payload = []
        values_payload_liq = []  # for liquefaction, if used
        unmapped_bridges = []
        mapped_bridges = []
        for b in bridges:
            bridge_id = b["id"]
            if bridge_id in fragility_set:
                location = GeoUtil.get_location(b)
                loc = str(location.y) + "," + str(location.x)

                demands = fragility_set[bridge_id].demand_types
                units = fragility_set[bridge_id].demand_units
                value = {
                    "demands": demands,
                    "units": units,
                    "loc": loc
                }
                values_payload.append(value)
                mapped_bridges.append(b)

                if use_liquefaction and geology_dataset_id is not None:
                    value_liq = {
                        "demands": [""],
                        "units": [""],
                        "loc": loc
                    }
                    values_payload_liq.append(value_liq)

            else:
                unmapped_bridges.append(b)

        # not needed anymore as they are already split into mapped and unmapped
        del bridges

        hazard_vals = hazard.read_hazard_values(values_payload, self.hazardsvc)

        # Check if liquefaction is applicable
        if use_liquefaction and geology_dataset_id is not None:
            liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id, geology_dataset_id,
                                                                        values_payload_liq)

        ds_results = []
        damage_results = []

        i = 0
        for bridge in mapped_bridges:
            ds_result = dict()
            damage_result = dict()
            dmg_probability = dict()
            dmg_intervals = dict()
            ground_failure_prob = None
            selected_fragility_set = fragility_set[bridge["id"]]

            if isinstance(selected_fragility_set.fragility_curves[0], DFR3Curve):
                # Supports multiple demand types in same fragility
                hazard_val = AnalysisUtil.update_precision_of_lists(hazard_vals[i]["hazardValues"])
                input_demand_types = hazard_vals[i]["demands"]
                input_demand_units = hazard_vals[i]["units"]

                hval_dict = dict()
                j = 0
                for d in selected_fragility_set.demand_types:
                    hval_dict[d] = hazard_val[j]
                    j += 1

                if not AnalysisUtil.do_hazard_values_have_errors(hazard_vals[i]["hazardValues"]):
                    bridge_args = selected_fragility_set.construct_expression_args_from_inventory(bridge)
                    dmg_probability = \
                        selected_fragility_set.calculate_limit_state(hval_dict,
                                                                     inventory_type="bridge",
                                                                     **bridge_args)

                    if use_liquefaction and geology_dataset_id is not None and liquefaction_resp is not None:
                        ground_failure_prob = liquefaction_resp[i][BridgeUtil.GROUND_FAILURE_PROB]

                        # TODO: put it in a util method
                        # Adjust for 4 LS of bridges - the service returns ground failure
                        # expecting 3 LS
                        ground_failure_prob.append(ground_failure_prob[2])
                        ground_failure_prob[2] = ground_failure_prob[1]

                        dmg_probability = AnalysisUtil.update_precision_of_dicts(
                            AnalysisUtil.adjust_damage_for_liquefaction(dmg_probability, ground_failure_prob))

                    dmg_intervals = selected_fragility_set.calculate_damage_interval(dmg_probability,
                                                                                     hazard_type=hazard_type,
                                                                                     inventory_type="bridge")
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            retrofit_cost = BridgeUtil.get_retrofit_cost(fragility_key)
            retrofit_type = BridgeUtil.get_retrofit_type(fragility_key)

            ds_result['guid'] = bridge['properties']['guid']
            ds_result.update(dmg_probability)
            ds_result.update(dmg_intervals)
            ds_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(hazard_val, hazard_type)

            damage_result['guid'] = bridge['properties']['guid']
            damage_result['fragility_id'] = selected_fragility_set.id
            damage_result["retrofit"] = retrofit_type
            damage_result["retrocost"] = retrofit_cost
            damage_result["demandtypes"] = input_demand_types
            damage_result["demandunits"] = input_demand_units
            damage_result["hazardtype"] = hazard_type
            damage_result["hazardval"] = hazard_val
            if use_liquefaction and geology_dataset_id is not None:
                damage_result[BridgeUtil.GROUND_FAILURE_PROB] = ground_failure_prob

            # add spans to bridge output so mean damage calculation can use that info
            if "spans" in bridge["properties"] and bridge["properties"]["spans"] is not None:
                if isinstance(bridge["properties"]["spans"], str) and bridge["properties"]["spans"].isdigit():
                    damage_result['spans'] = int(bridge["properties"]["spans"])
                elif isinstance(bridge["properties"]["spans"], int):
                    damage_result['spans'] = bridge["properties"]["spans"]
            elif "SPANS" in bridge["properties"] and bridge["properties"]["SPANS"] is not None:
                if isinstance(bridge["properties"]["SPANS"], str) and bridge["properties"]["SPANS"].isdigit():
                    damage_result['SPANS'] = int(bridge["properties"]["SPANS"])
                elif isinstance(bridge["properties"]["SPANS"], int):
                    damage_result['SPANS'] = bridge["properties"]["SPANS"]
            else:
                damage_result['spans'] = 1

            ds_results.append(ds_result)
            damage_results.append(damage_result)
            i += 1

        for bridge in unmapped_bridges:
            ds_result = dict()
            damage_result = dict()

            ds_result['guid'] = bridge['properties']['guid']

            damage_result['guid'] = bridge['properties']['guid']
            damage_result["retrofit"] = None
            damage_result["retrocost"] = None
            damage_result["demandtypes"] = None
            damage_result['demandunits'] = None
            damage_result["hazardtype"] = None
            damage_result['hazardval'] = None
            damage_result['spans'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        return ds_results, damage_results

    def get_spec(self):
        """Get specifications of the bridge damage analysis.

        Returns:
            obj: A JSON object of specifications of the bridge damage analysis.

        """
        return {
            'name': 'bridge-damage',
            'description': 'bridge damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use liquefaction',
                    'type': bool
                },
                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Geology dataset id',
                    'type': str,
                },
                {
                    'id': 'use_hazard_uncertainty',
                    'required': False,
                    'description': 'Use hazard uncertainty',
                    'type': bool
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard object id',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazards type',
                    'type': str
                },
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tornado", "hurricane", "flood", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'bridges',
                    'required': True,
                    'description': 'Bridge Inventory',
                    'type': ['ergo:bridges', 'ergo:bridgesVer2', 'ergo:bridgesVer3'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'bridges',
                    'description': 'CSV file of bridge structural damage',
                    'type': 'ergo:bridgeDamageVer3'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'bridges',
                    'description': 'additional metadata in json file about applied hazard value and '
                                   'fragility',
                    'type': 'incore:bridgeDamageSupplement'
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from pyincore.analyses.indp.infrastructurearc import InfrastructureArc


class InfrastructureInterdepArc(InfrastructureArc):
    """
    This class models a physical interdependency between nodes from two different layers. This class inherits from
    :class:`InfrastructureArc`, where `source` attributes corresponds to the dependee node, and `dest` corresponds to
    the depender node. The depender node is non-functional if the corresponding dependee node is non-functional.

    Attributes
    ----------
    source_layer : int
        The id of the layer where the dependee node is
    dest_layer : int
        The id of the layer where the depender node is
    gamma : float
        The strength of the dependency, which is a number between 0 and 1.
    """

    def __init__(self, source, dest, source_layer, dest_layer, gamma):
        super(InfrastructureInterdepArc, self).__init__(source, dest, source_layer, True)
        self.source_layer = source_layer
        self.dest_layer = dest_layer
        self.gamma = gamma



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class InfrastructureNode(object):
    """
    This class models a node in an infrastructure network

    Attributes
    ----------
    id : int
        Node id
    net_id : int
        The id of the layer of the network to which the node belong
    local_id : int
        Local node id
    failure_probability : float
        Failure probability of the node
    functionality : bool
        Functionality state of the node
    repaired : bool
        If the node is repaired or not
    reconstruction_cost : float
        Reconstruction cost of the node
    oversupply_penalty : float
        Penalty per supply unit  of the main commodity that is not used for the the node
    undersupply_penalty : float
        Penalty per demand unit  of the main commodity that is not satisfied for the the node
    demand : float
        Demand or supply value of the main commodity assigned to the node
    space : int
        The id of the geographical space where the node is
    resource_usage : dict
        The dictionary that shows how many resource (of each resource type) is employed to repair the node
    extra_com : dict
        The dictionary that shows demand, oversupply_penalty, and undersupply_penalty corresponding to commodities
        other than the main commodity
    """

    def __init__(self, id, net_id, local_id=""):
        self.id = id
        self.net_id = net_id
        if local_id == "":
            self.local_id = id
        self.local_id = local_id
        self.failure_probability = 0.0
        self.functionality = 1.0
        self.repaired = 1.0
        self.reconstruction_cost = 0.0
        self.oversupply_penalty = 0.0
        self.undersupply_penalty = 0.0
        self.demand = 0.0
        self.space = 0
        self.resource_usage = {}
        self.extra_com = {}

    def set_failure_probability(self, failure_probability):
        """
        This function sets the failure probability of the node

        Parameters
        ----------
        failure_probability : float
            Assigned failure probability of node

        Returns
        -------
        None.

        """
        self.failure_probability = failure_probability

    def set_extra_commodity(self, extra_commodity):
        """
        This function initialize the dictionary for the extra commodities

        Parameters
        ----------
        extra_commodity : list
            List of extra commodities

        Returns
        -------
        None.

        """
        for l in extra_commodity:
            self.extra_com[l] = {'demand': 0, 'oversupply_penalty': 0, 'undersupply_penalty': 0}

    def set_resource_usage(self, resource_names):
        """
        This function initialize the dictionary for resource usage per all resource types in the analysis

        Parameters
        ----------
        resource_names : list
            List of resource types

        Returns
        -------
        None.

        """
        for rc in resource_names:
            self.resource_usage[rc] = 0

    def in_space(self, space_id):
        """
        This function checks if the node is in a given space or not

        Parameters
        ----------
        space_id :
            The id of the space that is checked

        Returns
        -------
            : bool
            Returns 1 if the node is in the space, and 0 otherwise.
        """
        if self.space == space_id:
            return 1
        else:
            return 0



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore import Dataset


class HazardDataset:
    def __init__(self, hazard_datasets_metadata):
        self.hazard_type = hazard_datasets_metadata["hazardType"] if "hazardType" in hazard_datasets_metadata else ""
        self.demand_type = hazard_datasets_metadata["demandType"] if "demandType" in hazard_datasets_metadata else ""
        self.demand_units = hazard_datasets_metadata["demandUnits"] \
            if "demandUnits" in hazard_datasets_metadata else ""
        self.dataset_id = hazard_datasets_metadata["datasetId"] if "datasetId" in hazard_datasets_metadata else ""
        self.dataset = None

        # default threshold value and unit if exist
        self.threshold_value = hazard_datasets_metadata["threshold"] if "threshold" in hazard_datasets_metadata \
            else None
        self.threshold_unit = self.demand_units

    def from_file(self, file_path, data_type):
        """Get hurricane dataset from the file.
        """
        self.dataset = Dataset.from_file(file_path, data_type)

    def set_threshold(self, threshold_value, threshold_unit):
        """
        Set threshold value for the hazard to determine exposure or not
        Args:
            threshold_value: threshold_value
            threshold_unit: threshold_unit
        Returns:

        """
        self.threshold_value = threshold_value
        self.threshold_unit = threshold_unit

    def from_data_service(self, data_service):
        """Get hazard dataset from the data service."""
        self.dataset = Dataset.from_data_service(self.dataset_id, data_service)


class HurricaneDataset(HazardDataset):
    def __init__(self, hazard_datasets_metadata):
        super().__init__(hazard_datasets_metadata)
        self.hurricane_parameters = hazard_datasets_metadata["hurricaneParameters"] \
            if "hurricaneParameters" in hazard_datasets_metadata else {}


class EarthquakeDataset(HazardDataset):
    def __init__(self, hazard_datasets_metadata):
        super().__init__(hazard_datasets_metadata)
        self.period = hazard_datasets_metadata["period"] if "period" in hazard_datasets_metadata else 0
        self.recurrence_interval = hazard_datasets_metadata["recurrenceInterval"] if "recurrenceInterval" in  \
                                                                                     hazard_datasets_metadata else 10000
        self.recurrence_unit = hazard_datasets_metadata["recurrenceUnit"] if "recurrenceUnit" in \
                                                                             hazard_datasets_metadata else "years"
        self.eq_parameters = hazard_datasets_metadata["eqParameters"] if "eqParameters" in hazard_datasets_metadata \
            else {}


class TsunamiDataset(HazardDataset):
    def __init__(self, hazard_datasets_metadata):
        super().__init__(hazard_datasets_metadata)
        self.recurrence_interval = hazard_datasets_metadata["recurrenceInterval"] \
            if "recurrenceInterval" in hazard_datasets_metadata else 100
        self.recurrence_unit = hazard_datasets_metadata["recurrenceUnit"] if "recurrenceUnit" in \
                                                                             hazard_datasets_metadata else "years"


# TODO: Tornado dataset has very different shape
class TornadoDataset(HazardDataset):
    def __init__(self, hazard_datasets_metadata):
        super().__init__(hazard_datasets_metadata)


class FloodDataset(HazardDataset):
    def __init__(self, hazard_datasets_metadata):
        super().__init__(hazard_datasets_metadata)
        self.flood_parameters = hazard_datasets_metadata["floodParameters"] \
            if "floodParameters" in hazard_datasets_metadata else {}




# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from copy import deepcopy
import pandas as pd
import operator as op
import math


class VarContainer:
    """
     All matrix variable(tables) in the GAMS model is flatten to a array to make a better
     interface to the solver.

     AllVarList stores all initial values of varibles used in the GAMS model in an array.
     It also has a indexing system for looking up.

     Attributes:
         namelist: A dictionary with all stored GAMS variables and its information.
         nvars: The length of the array, i.e. the size of all matrix variables summed up.
         initialVals: Stored initial values of all variables

    """

    def __init__(self):
        """
          Initialize to an empty list
        """
        self.namelist = {}
        self.initialVals = []
        self.nvars = 0
        self.LO = []
        self.UP = []

    def add(self, name, rows=None, cols=None):
        """
        :param name:
        :param rows:
        :param cols:
        :return:
        """
        if rows is not None and cols is not None:
            size = len(rows) * len(cols)
            self.namelist[name] = { \
                'nrows': len(rows), \
                'ncols': len(cols), \
                'rows': rows, \
                'cols': cols, \
                'start': self.nvars, \
                'size': size
            }
        elif rows is not None and cols is None:
            size = len(rows)
            self.namelist[name] = {
                'nrows': len(rows), \
                'rows': rows, \
                'start': self.nvars, \
                'size': size
            }
        else:
            size = 1
            self.namelist[name] = {'start': self.nvars, 'size': 1}

        self.initialVals.extend([None] * size)
        self.LO.extend([None] * size)
        self.UP.extend([None] * size)
        self.nvars += size

        return ExprM(self, name=name, rows=rows, cols=cols)

    def set_value(self, name, values, target):
        """
        An internal method for setting the initial values or UPs and LOs for variables

        :param name: Name of the variable in GAMS
        :param value: a pandas DataFrame, pandas Series, int or float with initial values
        :param target: target array to be set

        :return: None
        """
        if type(values) == int or type(values) == float:
            info = self.namelist[name]
            if 'nrows' in info and 'ncols' in info:
                values = pd.DataFrame(index=info['rows'], columns=info['cols']).fillna(values)
            elif 'nrows' in info and 'ncols' not in info:
                values = pd.Series(index=info['rows'], dtype="float64").fillna(values)

        if type(values) == pd.DataFrame:
            rows = values.index.tolist()
            cols = values.columns.tolist()
            for i in rows:
                for j in cols:
                    target[self.getIndex(name, row=i, col=j)] = float(values.loc[i][j])
        elif type(values) == pd.Series:
            rows = values.index.tolist()
            for i in rows:
                target[self.getIndex(name, row=i)] = float(values.loc[i])
        else:
            target[self.getIndex(name)] = values

    def init(self, name, initialValue):
        """
        Flatten the table variable and add to the list.
        Also set the initial variable values array.

        :param name: Name of the variable in GAMS
        :param initialValue: a pandas DataFrame or pandas Series with initial values

        :return: None.
        """
        self.set_value(name, initialValue, self.initialVals)

    def inList(self, name):
        """
        Check if a GAMS varible is added to the container

        :param name(str): name of GAMS variable you want to look up
        :return: Boolean, whether the variable is added.
        """
        return name in self.namelist

    def getInfo(self, name):
        """
        Get the information about a GAMS variable

        :param name(str): name of GAMS variable you want to look up
        :return: a dictionary with all information
        """
        return self.namelist[name]

    def getIndex(self, name, row=None, col=None):
        """
        Look up the index by providing the variable name and label information

        :param name: name of GAMS variable you want to look up
        :param row: row label of the position you want to look up index for(if it has row labels)
        :param col: column label of the position you want to look up index for(if it has column labels)
        :return: the index of the position in the array
        """
        info = self.namelist[name]
        result = info['start']
        if row is not None and col is not None:
            result += info['rows'].index(row) * info['ncols'] + info['cols'].index(col)
        elif row is not None:
            result += info['rows'].index(row)
        return result

    def getLabel(self, index):
        """
        Look up variable name and label information by providing the index

        :param index: the index in the array
        :return: its information including the variable name, row label and column label if applicable
        """
        result = []
        for i in self.namelist.keys():
            if index >= self.namelist[i]['start'] and index < self.namelist[i]['start'] + self.namelist[i]['size']:
                result.append(i)
                if self.namelist[i]['size'] > 1:
                    diff = index - self.namelist[i]['start']
                    if 'ncols' in self.namelist[i]:
                        result.append(self.namelist[i]['rows'][int(diff / self.namelist[i]['ncols'])])
                        result.append(self.namelist[i]['cols'][diff % self.namelist[i]['ncols']])
                    else:
                        result.append(self.namelist[i]['rows'][diff])
                return result

    def get_all_variable_printed(self, output=None):
        if output is None:
            output = self.initialVals
        for i in range(len(output)):
            label = self.getLabel(i)
            if len(label) == 1:
                print(label[0] + '=' + '{0:.7f}'.format(output[i]) + ';')
            elif len(label) == 2:
                print(label[0] + '(\'' + label[1] + '\')' + '=' + '{0:.7f}'.format(output[i]) + ';')
            elif len(label) == 3:
                print(
                    label[0] + '(\'' + label[1] + '\',\'' + label[2] + '\')' + '=' + '{0:.7f}'.format(output[i]) + ';')

    def get(self, name, x=None):
        """
          Returns a Dataframe, Series, or a variable based on the given name and the result array returned from the solver

          :param name: GAMS variable name
          :return: if x is not given, it returns the initial values
            if x is set to the result, returns the result variable value
        """
        if x is None:
            x = self.initialVals

        info = self.namelist[name]
        if 'nrows' in info and 'ncols' in info:
            ret = pd.DataFrame(index=info['rows'], columns=info['cols']).fillna(0.0)
            for i in info['rows']:
                for j in info['cols']:
                    ret.at[i, j] = x[self.getIndex(name, row=i, col=j)]
        elif 'nrows' in info and 'ncols' not in info:
            ret = pd.Series(index=info['rows'], dtype="float64").fillna(0.0)
            for i in info['rows']:
                ret.at[i] = x[self.getIndex(name, row=i)]
        elif 'nrows' not in info and 'ncols' not in info:
            ret = x[self.getIndex(name)]

        return ret

    def lo(self, name, value):
        """
          Set the LOs of a GAMS variable providing the LOs with a Dataframe, Series, int or float

          :param name: GAMS variable name
          :param value: The lower bound to be set
          :return: None
        """
        self.set_value(name, value, self.LO)

    def up(self, name, value):
        """
          Set the UPs of a GAMS variable providing the LOs with a Dataframe, Series, int or float

          :param name: GAMS variable name
          :param value: The upper bound to be set
          :return: None
        """
        self.set_value(name, value, self.UP)

    def write(self, filename):
        """
          Write(append) the variables to a file, in the format of setting ipopt model variables

          :param filename: the output filename
          :return: None
        """
        with open(filename, 'a') as f:
            for i in range(self.nvars):
                lower = -1e20 if self.LO[i] == None else self.LO[i]
                upper = 1e20 if self.UP[i] == None else self.UP[i]
                value = 0 if math.isnan(self.initialVals[i]) else self.initialVals[i]

                # tmp_list = ['CG','CH','CN','CPI','CPIN','CPIH','CX','D','DD','DS','FD','GCP','GCP1','HH','HN','HW','IGT','KS']
                '''tmp_list = ['CG','CH','CN','CPI','CPIN','CPIH','CX','D','DD','DS','FD','GCP','GCP1','HH','HN','HW','IGT','KS','LAS','M','N','NKI','LNFOR','KPFOR','GVFOR','P','PD','PVA']                        
                if self.getLabel(i)[0] in tmp_list:
                    f.write('model.x' + str(i) + ' = Var(bounds=(' + str(lower) + ',' + str(upper) + '),initialize=' + str(self.initialVals[i]) + ')' + '\n')
                else:
                    f.write('model.x' + str(i) + ' = Var(bounds=(' + str(self.initialVals[i]-1e-3) + ',' + str(self.initialVals[i]+1e-3) + '),initialize=' + str(self.initialVals[i]) + ')' + '\n')
                '''
                # if i == 1507:
                #  f.write('model.x' + str(i) + ' = Var(bounds=(' + str(value) + ',' + str(value*10) + '),initialize=' + str(value) + ')' + '\n')
                # else:
                #  f.write('model.x' + str(i) + ' = Var(bounds=(' + str(lower) + ',' + str(upper) + '),initialize=' + str(value) + ')' + '\n')
                f.write('model.x' + str(i) + ' = Var(bounds=(' + str(lower) + ',' + str(upper) + '),initialize=' + str(
                    value) + ')' + '\n')


class variable:
    """
      A GMAS variable, initialized by given the GAMS variable value and its label
    """

    def __init__(self, vars, name, row=None, col=None):
        """
          Initialize it with a variable container, the GAMS name, the labels

          :param vars: the variable container that already added the GAMS variable
          :param name: GAMS variable name
          :param row: GAMS row label if there is
          :param col: GAMS col label if there is
          :return: None
        """
        try:
            self.index = vars.getIndex(name, row, col)
        except Exception as e:
            print(e)
            # print("invalid name for a variable")

    def __str__(self):
        """
          returns the variable in the format of "model.x#" if gets printed,
          with # being the index in the array in the container

          :return: String
        """
        return 'model.x' + str(self.index) + ''
        # return 'x[' + str(self.index) + ']'

    def debug_test_str(self):
        return 'x[' + str(self.index) + ']'


class ExprItem:
    '''
      You can construct it with a variable, a constant or a deepcopy of another ExprItem
    '''

    def __init__(self, v, const=1):
        self.varList = []
        self.const = const
        if type(v) == variable:
            self.varList.append(v)
        elif type(v) == int or type(v) == float:
            self.const = v
        elif type(v) == ExprItem:
            self.const = deepcopy(v.const)
            self.varList = deepcopy(v.varList)
        else:
            print("invalid parameter to create a item")

    '''
      You could multiply it with a number, a variable, a ExprItem or Expression
    '''

    def __mul__(self, rhs):
        copy = ExprItem(self)
        if type(rhs) == int or type(rhs) == float:
            copy.const = copy.const * rhs
        elif type(rhs) == variable:
            copy.varList.append(rhs)
        elif type(rhs) == ExprItem:
            copy.const *= rhs.const
            copy.varList.extend(deepcopy(rhs.varList))
        elif type(rhs) == Expr:
            if rhs.isComposite:
                if rhs.operator == '/':
                    return rhs * copy
                else:
                    copy.varList.append(Expr(rhs))
            else:
                copyrhs = Expr(rhs)
                copyrhs.itemList = [i * copy for i in copyrhs.itemList]
                return copyrhs
        return copy

    def __str__(self):
        # if abs(self.const) > 0.1:
        #  result = '' + str(round(self.const,3))
        # else:
        #  result = '' + str(self.const)
        # result = '' + str(round(self.const,7))
        result = '' + str(self.const)
        for i in range(len(self.varList)):
            result += "*" + self.varList[i].__str__()
        return result

    def debug_test_str(self):
        # if abs(self.const) > 0.1:
        #  result = '' + str(round(self.const,3))
        # else:
        #  result = '' + str(self.const)
        result = '' + str(self.const)
        for i in range(len(self.varList)):
            result += "*" + self.varList[i].debug_test_str()
        return result

    def is_empty(self):
        if abs(self.const) < 0.00000001:
            return True
        else:
            return False

    def is_one(self):
        if abs(self.const - 1) < 0.0000001 and len(self.varList) == 0:
            return True
        else:
            return False


class Expr:

    def __init__(self, item):
        self.itemList = []
        self.isComposite = False
        if type(item) == ExprItem or type(item) == variable or type(item) == int or type(item) == float:
            self.itemList.append(ExprItem(item))
        elif type(item) == Expr:
            self.itemList = [
                ExprItem(item.itemList[i]) if type(item.itemList[i]) == ExprItem else Expr(item.itemList[i]) for i in
                range(len(item.itemList))]
            try:
                self.isComposite = deepcopy(item.isComposite)
            except Exception as e:
                pass
            try:
                self.operator = deepcopy(item.operator)
            except Exception as e:
                pass
            try:
                self.first = Expr(item.first)
            except Exception as e:
                pass
            try:
                self.second = Expr(item.second)
            except Exception as e:
                pass
        else:
            print("invalid parameter for creating a Expr")

        self.clear_empty()

    def __add__(self, rhs):
        copy = Expr(self)
        if self.isComposite:
            tmp = Expr(self)
            copy.itemList = [tmp]
            copy.isComposite = False
        if type(rhs) == Expr:
            if rhs.isComposite:
                rhscopy = Expr(rhs)
                copy.itemList.append(rhscopy)
            else:
                rhscopy = Expr(rhs)
                copy.itemList = copy.itemList + rhscopy.itemList
        elif type(rhs) == ExprItem or type(rhs) == int or type(rhs) == float or type(rhs) == variable:
            copy.itemList.append(ExprItem(rhs))
        return copy

    def __sub__(self, rhs):
        copy = Expr(self)
        return copy + rhs * -1

    def __mul__(self, rhs):
        copy = Expr(self)
        if type(rhs) == int or type(rhs) == float or type(rhs) == ExprItem:
            if copy.isComposite:
                if copy.operator == '/':
                    copy.first = copy.first * rhs
                elif copy.opeartor == '**':
                    return ExprItem(rhs) * copy
            else:
                result = []
                for i in copy.itemList:
                    result.append(i * rhs)
                copy.itemList = result


        elif type(rhs) == Expr:
            if copy.isComposite and not rhs.isComposite:
                if copy.operator == '/':
                    copy.first = copy.first * rhs
                else:
                    copy.itemList = [i * copy for i in rhs.itemList]
                    copy.isComposite = False
            elif not copy.isComposite and not rhs.isComposite:
                # print('copy',copy)
                # print('rhs',rhs)

                if len(copy.itemList) * len(rhs.itemList) > 10:

                    tmpItem = ExprItem(1)
                    tmpItem.varList.append(Expr(copy))
                    tmpItem.varList.append(Expr(rhs))
                    copy.itemList = [tmpItem]

                else:

                    # The expanding way is super slow for long equations

                    result = []
                    for i in copy.itemList:
                        for j in rhs.itemList:
                            if type(i) == Expr:
                                tmp = Expr(i)
                            elif type(i) == ExprItem:
                                tmp = ExprItem(i)
                            result.append(tmp * j)
                    copy.itemList = result

            elif not self.isComposite and rhs.isComposite:
                return rhs * copy
            else:
                # both are composite
                if copy.operator == '/' and rhs.operator == '/':
                    copy.first = copy.first * rhs.first
                    copy.second = copy.second * rhs.second
                elif copy.operator == '**' and rhs.operator == '**':
                    copy.first = copy.first * rhs.first
                    copy.second = copy.second + rhs.second
                elif copy.operator == '/' and rhs.operator == '**':
                    copy.first = copy.first * rhs
                elif copy.operator == '**' and rhs.operator == '/':
                    copyrhs = Expr(rhs)
                    copyrhs.first = copyrhs.first * copy
        return copy

    def __truediv__(self, rhs):
        copy = Expr(self)
        tmp = Expr(copy)
        copy.isComposite = True
        copy.operator = '/'
        copy.first = tmp
        copy.second = Expr(rhs)
        return copy

    def __pow__(self, rhs):
        copy = Expr(self)
        tmp = Expr(copy)
        copy.isComposite = True
        copy.operator = '**'
        copy.first = tmp
        copy.second = Expr(rhs)
        return copy

    def __str__(self):
        if self.isComposite:
            result = "(" + self.first.__str__() + self.operator + self.second.__str__() + ")"
        else:
            if len(self.itemList) == 0:
                return '0'
            result = ''
            for i in self.itemList[0:-1]:
                result += "(" + i.__str__() + ")" + "+"
            result += "(" + self.itemList[-1].__str__() + ")"
            if len(self.itemList) > 1:
                result = "(" + result + ")"
        return result

    def debug_test_str(self):
        if self.isComposite:
            result = "(" + self.first.debug_test_str() + self.operator + self.second.debug_test_str() + ")"
        else:
            if len(self.itemList) == 0:
                return '0'
            result = ''
            for i in self.itemList[0:-1]:
                result += "(" + i.debug_test_str() + ")" + "+"
            result += "(" + self.itemList[-1].debug_test_str() + ")"
            if len(self.itemList) > 1:
                result = "(" + result + ")"
        return result

    def is_empty(self):
        if not self.isComposite and len(self.itemList) == 0:
            return True
        else:
            return False

    def clear_empty(self):
        result = []
        if self.isComposite == False:
            for i in range(len(self.itemList)):
                if type(self.itemList[i]) == ExprItem:
                    if not self.itemList[i].is_empty():
                        result.append(self.itemList[i])
                if type(self.itemList[i]) == Expr:
                    self.itemList[i].clear_empty()
                    if not self.itemList[i].is_empty():
                        result.append(self.itemList[i])
            self.itemList = result
        """
        else:
          if self.operator == '**' and is_empty(self.second):
            self.isComposite = False
            self.itemList = []
            self.itemList.append(ExprItem(1))
        """


class ExprM:
    '''
    Three ways to create a ExprMatrix:
    1. Give it the variable name, selected rows and cols(could be empty),
      The constructor will create a Expression matrix from the variable matrix
    2. Give it a pandas Series or DataFrame, it will create the Expression matrix
      with the content in the Series or DataFrame as constants
    3. Give it a ExprMatrix, will return a deep copy of it
    '''

    def __init__(self, vars, name=None, rows=None, cols=None, m=None, em=None):
        self.vars = vars
        self.hasCondition = False
        if em is None:
            # if these are the variables, we need to create an Expression by the variable name
            if name is not None:
                if self.vars.inList(name):
                    self.info = deepcopy(self.vars.getInfo(name))
                else:
                    print("Can't find this variable in the all variable list")

                self.info['height'] = 1
                self.info['width'] = 1
                self.info['rows'] = deepcopy(rows)
                self.info['cols'] = deepcopy(cols)
                if cols is not None:
                    self.info['width'] = len(cols)
                if rows is not None:
                    self.info['height'] = len(rows)

                if cols is not None:  # if it is a DataFrame
                    self.m = [[Expr(variable(self.vars, name, i, j)) for j in cols] for i in rows]
                elif rows is not None:  # if it is a Series
                    self.m = [[Expr(variable(self.vars, name, i))] for i in rows]
                else:  # if it is a variable
                    self.m = [[Expr(variable(self.vars, name))]]

            # otherwise these are just constants
            else:
                self.info = {}
                self.info['height'] = 1
                self.info['width'] = 1
                self.info['rows'] = None
                self.info['cols'] = None
                if type(m) == pd.DataFrame:
                    self.info['rows'] = m.index.tolist()
                    self.info['height'] = len(self.info['rows'])
                    self.info['cols'] = m.columns.tolist()
                    self.info['width'] = len(self.info['cols'])
                    self.m = [[Expr(float(m.loc[i, j])) for j in self.info['cols']] for i in self.info['rows']]
                elif type(m) == pd.Series:
                    self.info['rows'] = m.index.tolist()
                    self.info['height'] = len(self.info['rows'])
                    self.m = [[Expr(float(m.loc[i]))] for i in self.info['rows']]
                else:
                    self.m = [[Expr(float(m))]]
        else:
            # make deep copy
            self.info = deepcopy(em.info)
            self.m = [[Expr(em.m[i][j]) for j in range(em.info['width'])] for i in range(em.info['height'])]

    def operation(self, rhs, oper):
        copy = ExprM(self.vars, em=self)

        if type(rhs) == int or type(rhs) == float:
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    copy.m[i][j] = oper(copy.m[i][j], Expr(rhs))

        elif type(rhs) == pd.DataFrame or type(rhs) == pd.Series:
            return self.operation(ExprM(self.vars, m=rhs), oper)

        elif type(rhs) == ExprM:
            if copy.info['rows'] == rhs.info['rows'] and copy.info['cols'] == rhs.info['cols']:
                # same size, apply operation element-wise
                for i in range(self.info['height']):
                    for j in range(self.info['width']):
                        copy.m[i][j] = oper(copy.m[i][j], rhs.m[i][j])
            elif copy.info['rows'] == rhs.info['rows'] and copy.info['width'] == 1:
                # for each column in rhs
                result = [[Expr(0) for j in range(rhs.info['width'])] for i in range(self.info['height'])]
                for i in range(self.info['height']):
                    for j in range(rhs.info['width']):
                        result[i][j] = oper(copy.m[i][0], rhs.m[i][j])
                copy.m = result
                copy.info['width'] = rhs.info['width']
                copy.info['cols'] = deepcopy(rhs.info['cols'])
            elif copy.info['rows'] == rhs.info['rows'] and rhs.info['width'] == 1:
                # for each column in lhs
                for i in range(self.info['height']):
                    for j in range(self.info['width']):
                        copy.m[i][j] = oper(copy.m[i][j], rhs.m[i][0])
            elif copy.info['height'] == 1 and copy.info['cols'] == rhs.info['cols']:
                # for each row in rhs
                result = [[Expr(0) for j in range(rhs.info['width'])] for i in range(rhs.info['height'])]
                for i in range(rhs.info['height']):
                    for j in range(rhs.info['width']):
                        result[i][j] = oper(copy.m[0][j], rhs.m[i][j])
                copy.m = result
                copy.info['height'] = rhs.info['height']
                copy.info['rows'] = deepcopy(rhs.info['rows'])
            elif rhs.info['height'] == 1 and copy.info['cols'] == rhs.info['cols']:
                # for each row in lhs
                for i in range(self.info['height']):
                    for j in range(self.info['width']):
                        copy.m[i][j] = oper(copy.m[i][j], rhs.m[0][j])
            elif (self.info['width'] == rhs.info['height'] and self.info['cols'] == rhs.info['rows']) or \
                    (self.info['height'] == rhs.info['width'] and self.info['rows'] == rhs.info['cols']):
                # flip the matrix
                return oper(self, ~rhs)

            else:
                print(copy.info['rows'], copy.info['cols'], rhs.info['rows'], rhs.info['cols'])
                print(copy.info['height'], copy.info['width'], rhs.info['height'], rhs.info['width'])
                print("Invalid size for ", str(oper))
        return copy

    def __add__(self, rhs):
        return self.operation(rhs, op.add)

    def __radd__(self, lhs):
        return self + lhs

    def __sub__(self, rhs):
        return self.operation(rhs, op.sub)

    def __rsub__(self, lhs):
        return self * -1 + lhs

    def __mul__(self, rhs):
        return self.operation(rhs, op.mul)

    def __truediv__(self, rhs):
        return self.operation(rhs, op.truediv)

    def __pow__(self, rhs):
        return self.operation(rhs, op.pow)

    def __xor__(self, rhs):
        """
          create 2d list out of 2 single lists
        """
        # has to be 2 single lists
        if self.info['width'] != 1 or rhs.info['width'] != 1:
            print("Invalid size for creating a 2-D matrix")
        else:
            copy = ExprM(self.vars, em=self)
            copy.m = [[copy.m[i][0] * rhs.m[j][0] for j in range(rhs.info['height'])] for i in
                      range(copy.info['height'])]
            copy.info['cols'] = deepcopy(rhs.info['rows'])
            copy.info['width'] = deepcopy(rhs.info['height'])
            return copy

    def __invert__(self):
        """
          Return the transpose of a Expression matrix
        """
        copy = ExprM(self.vars, em=self)
        result = [[copy.m[i][j] for i in range(copy.info['height'])] for j in range(copy.info['width'])]
        copy.info['height'], copy.info['width'] = copy.info['width'], copy.info['height']
        copy.info['rows'], copy.info['cols'] = copy.info['cols'], copy.info['rows']
        copy.m = result
        return copy

    def __str__(self):
        result = ''
        for i in self.m:
            for j in i:
                result += j.__str__() + '\n'
            result += '///////////////////\n'
        return result

    def loc(self, rows=None, cols=None):
        """
          get a subset of the matrix by labels
        """
        copy = ExprM(self.vars, em=self)
        if cols is not None:
            result = [[Expr(copy.m[self.info['rows'].index(i)][self.info['cols'].index(j)]) for j in cols] for i in
                      rows]
            copy.m = result
            copy.info['rows'] = deepcopy(rows)
            copy.info['height'] = len(copy.info['rows'])
            copy.info['cols'] = deepcopy(cols)
            copy.info['width'] = len(copy.info['cols'])
            return copy
        elif rows is not None:
            result = [[Expr(copy.m[self.info['rows'].index(i)][0])] for i in rows]
            copy.m = result
            copy.info['rows'] = deepcopy(rows)
            copy.info['height'] = len(copy.info['rows'])
            return copy
        else:
            return copy

    def sum(self, label=None):
        copy = ExprM(self.vars, em=self)
        if label == None:
            result = Expr(0)
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    result = result + copy.m[i][j]
            copy.m = [[result]]
            copy.info['width'] = 1
            copy.info['height'] = 1
            copy.info['rows'] = None
            copy.info['cols'] = None

        elif label == self.info['rows'] or label == 0:
            result = [[Expr(0) for j in range(self.info['width'])]]
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                        result[0][j] = result[0][j] + copy.m[i][j]
            copy.m = result
            copy.info['height'] = 1
            copy.info['rows'] = None

        elif label == self.info['cols'] or label == 1:
            result = [[Expr(0)] for i in range(self.info['height'])]
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                        result[i][0] = result[i][0] + copy.m[i][j]
            copy.m = result
            copy.info['width'] = 1
            copy.info['cols'] = None
        return copy

    def prod(self, label):
        copy = ExprM(self.vars, em=self)
        if label == self.info['rows'] or label == 0:
            result = [[Expr(1) for j in range(self.info['width'])]]
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                        result[0][j] = result[0][j] * copy.m[i][j]
            copy.m = result
            copy.info['height'] = 1
            copy.info['rows'] = None
        elif label == self.info['cols'] or label == 1:
            result = [[Expr(1)] for i in range(self.info['height'])]
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                        result[i][0] = result[i][0] * copy.m[i][j]
            copy.m = result
            copy.info['width'] = 1
            copy.info['cols'] = None
        return copy

    def setCondition(self, matrix, operator=None, value=None):
        self.hasCondition = True
        mappings = {'LT': op.lt, 'LE': op.le, 'EQ': op.eq, 'INEQ': op.ne}  # not complete list
        if type(matrix) == pd.DataFrame:
            self.mark = [[False for j in range(self.info['width'])] for i in range(self.info['height'])]
            for i in range(self.info['height']):
                for j in range(self.info['width']):
                    if operator == None:
                        self.mark[i][j] = op.ne(matrix.loc[self.info['rows'][i]][self.info['cols'][j]], 0)
                    else:
                        self.mark[i][j] = mappings[operator](matrix.loc[self.info['rows'][i]][self.info['cols'][j]],
                                                             value)

        if type(matrix) == pd.Series:
            self.mark = [[False] for i in range(self.info['height'])]
            for i in range(self.info['height']):
                if operator == None:
                    self.mark[i][0] = op.ne(matrix.loc[self.info['rows'][i]], 0)
                else:
                    self.mark[i][0] = mappings[operator](matrix.loc[self.info['rows'][i]], value)

    def write(self, count, filename):
        f = open(filename, 'a')
        for i in range(self.info['height']):
            for j in range(self.info['width']):
                if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                    f.write(
                        'model.equality' + str(count[0]) + ' = Constraint(expr=' + self.m[i][j].__str__() + ' == 0)\n')
                    count[0] += 1
        f.close()

    def test(self, x):
        for i in range(self.info['height']):
            for j in range(self.info['width']):
                if not self.hasCondition or self.hasCondition and self.mark[i][j]:
                    fun = lambda x: eval(self.m[i][j].debug_test_str())
                    print(i, j, fun(x))



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore import BaseAnalysis, AnalysisUtil

import numpy as np
import pandas as pd
import concurrent.futures
from itertools import repeat


class HousingRecoverySequential(BaseAnalysis):
    """
    This analysis computes the series of household recovery states given a population
    dislocation dataset, a transition probability matrix (TPM) and an initial state vector.

    The computation operates by segregating household units into five zones as a way of
    assigning social vulnerability. Using this vulnerability in conjunction with the TPM
    and the initial state vector, a Markov chain computation simulates most probable
    states to generate a stage history of housing recovery changes for each household.

    The output of the computation is the history of housing recovery changes for each household unit in CSV format.

    Contributors
        | Science: Elaina Sutley, Sara Hamideh
        | Implementation: Nathanael Rosenheim, Santiago Núñez-Corrales, and NCSA IN-CORE Dev Team

    Related publications
        Sutley, E.J. and Hamideh, S., 2020. Postdisaster housing stages: a Markov chain approach to model sequences
        and duration based on social vulnerability. Risk Analysis, 40(12), pp.2675-2695.

    Args:
        incore_client (IncoreClient): Service authentication.
    """

    # Social vulnerability value generators per zone
    __sv_generator = {
        'Z1': {
            'threshold_0': 0.95,
            'below_lower': 0.00,
            'below_upper': 0.20,
            'above_lower': 0.20,
            'above_upper': 1.00
        },
        'Z2': {
            'threshold_0': 0.85,
            'below_lower': 0.20,
            'below_upper': 0.40,
            'threshold_1': 0.90,
            'middle_lower': 0.00,
            'middle_upper': 0.20,
            'above_lower': 0.40,
            'above_upper': 1.00
        },
        'Z3': {
            'threshold_0': 0.80,
            'below_lower': 0.40,
            'below_upper': 0.60,
            'threshold_1': 0.90,
            'middle_lower': 0.00,
            'middle_upper': 0.40,
            'above_lower': 0.60,
            'above_upper': 1.00
        },
        'Z4': {
            'threshold_0': 0.85,
            'below_lower': 0.60,
            'below_upper': 0.80,
            'threshold_1': 0.95,
            'middle_lower': 0.00,
            'middle_upper': 0.40,
            'above_lower': 0.80,
            'above_upper': 1.00
        },
        'Z5': {
            'threshold_0': 0.95,
            'below_lower': 0.80,
            'below_upper': 1.00,
            'above_lower': 0.00,
            'above_upper': 0.80
        }
    }

    def __init__(self, incore_client):
        super(HousingRecoverySequential, self).__init__(incore_client)

    def run(self):
        """Execute the HHRS analysis using parameters and input data."""
        # Read parameters
        t_delta = self.get_parameter('t_delta')
        t_final = self.get_parameter('t_final')

        # Load population block data from IN-CORE
        pop_dis_selectors = ['guid', 'huid', 'blockid', 'race', 'hispan', 'ownershp', 'dislocated']
        households_csv = self.get_input_dataset('population_dislocation_block').get_csv_reader()
        households_df = (pd.DataFrame(households_csv))[pop_dis_selectors]

        # Perform  conversions across the dataset from object type into the appropriate type
        households_df['blockid'] = households_df['blockid'].astype('int64')
        households_df['race'] = pd.to_numeric(households_df['race'])
        households_df['hispan'] = pd.to_numeric(households_df['hispan'])
        households_df['ownershp'] = pd.to_numeric(households_df['ownershp'])
        households_df['dislocated'] = (households_df['dislocated'] == 'True')

        # Load the transition probability matrix from IN-CORE
        tpm_csv = self.get_input_dataset("tpm").get_csv_reader()
        tpm = pd.DataFrame(list(tpm_csv))

        # Make sure all columns have numeric values
        for c_name in list(tpm.columns.values):
            tpm[c_name] = pd.to_numeric(tpm[c_name])

        tpm = tpm.to_numpy()
        tpm[:, 0] = np.around(tpm[:, 0], 3)

        # Load the initial stage probability vector
        initial_prob_csv = self.get_input_dataset("initial_stage_probabilities").get_csv_reader()
        initial_prob = pd.DataFrame(list(initial_prob_csv))
        initial_prob['value'] = pd.to_numeric(initial_prob['value'])

        # Obtain the number of CPUs (cores) to execute the analysis with
        user_defined_cpu = 4

        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(households_df), user_defined_cpu)

        # Chop dataset into `num_size` chunks
        max_chunk_size = int(np.ceil(len(households_df)/num_workers))

        households_df_list = list(
            map(lambda x: households_df[x * max_chunk_size:x * max_chunk_size + max_chunk_size],
                list(range(num_workers)))
        )

        # Run the analysis
        result = self.hhrs_concurrent_future(self.housing_serial_recovery_model,
                                                 num_workers,
                                                 households_df_list,
                                                 repeat(t_delta),
                                                 repeat(t_final),
                                                 repeat(tpm),
                                                 repeat(initial_prob))

        result_name = self.get_parameter("result_name")
        self.set_result_csv_data("ds_result", result, name=result_name, source="dataframe")

        return True

    def hhrs_concurrent_future(self, function_name, parallelism, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            parallelism (int): Number of workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            pd.DataFrame: outcome DataFrame containing the results from the concurrent function.

        """
        output_ds = pd.DataFrame()

        with concurrent.futures.ProcessPoolExecutor(max_workers=parallelism) as executor:
            for ret in executor.map(function_name, *args):
                output_ds = pd.concat([output_ds, ret], ignore_index=True)

        return output_ds

    def housing_serial_recovery_model(self, households_df, t_delta, t_final, tpm, initial_prob):
        """Performs the computation of the model as indicated in Sutley and Hamide (2020).

        Args:
            households_df (pd.DataFrame): Households with population dislocation data.
            t_delta (float): Time step size.
            t_final (float): Final time.
            tpm (np.Array): Transition probability matrix.
            initial_prob (pd.DataFrame): Initial probability Markov vector.

        Returns:
            pd.DataFrame: outcome of the HHRS model for a given household dataset.
        """
        seed = self.get_parameter('seed')
        rng = np.random.RandomState(seed)
        sv_result = self.get_input_dataset("sv_result").get_dataframe_from_csv(low_memory=False)
        # turn fips code to string for ease of matching
        sv_result["FIPS"] = sv_result["FIPS"].astype(str)

        # Compute the social vulnerability zone using known factors
        households_df = self.compute_social_vulnerability_zones(sv_result, households_df)

        # Set the number of Markov chain stages
        stages = int(t_final / t_delta)

        # Data structure for selection operations
        initial_prob['cumulative'] = initial_prob['value'].cumsum()

        # Obtain number of households with social vulnerability zones
        num_households = households_df.shape[0]

        # Obtain a social vulnerability score stochastically per household
        # We use them later to construct the final output dataset
        sv_scores = self.compute_social_vulnerability_values(households_df, num_households, rng)

        # We store Markov states as a list of numpy arrays for convenience and add each one by one
        markov_stages = np.zeros((stages, num_households))

        for household in range(0, num_households):
            if households_df['dislocated'].iat[household]:
                spin = rng.rand()

                if spin < initial_prob['cumulative'][0]:
                    markov_stages[0][household] = 1.0
                elif spin < initial_prob['cumulative'][1]:
                    markov_stages[0][household] = 2.0
                else:
                    markov_stages[0][household] = 3.0
            else:
                markov_stages[0][household] = 4.0

        # With all data in place, run the Markov chain (Cell 36 onwards)

        # Iterate over all stages
        for t in range(1, stages):
            # Generate a vector with random numbers for all households
            spins = rng.rand(num_households)

            # Iterate over all households
            for household in range(0, num_households):
                stage_val = markov_stages[t - 1][household]

                if stage_val == 5:
                    # If the household has already transitioned to stage 5, the household will remain in stage 5.
                    markov_stages[t][household] = 5
                else:
                    # Obtain the vector used to setup the stage
                    mv_index = np.where(tpm[:, 0] == sv_scores[household])[0][0]
                    lower = int((stage_val - 1) * 4 + 1)
                    upper = int(stage_val * 4 + 1)
                    mv = tpm[mv_index, lower:upper]
                    markov_vector = [mv[0], sum(mv[0:2]), sum(mv[0:3])]

                    # Set the new stage using the Markov property
                    if spins[household] < markov_vector[0]:
                        markov_stages[t][household] = 1
                    elif spins[household] < markov_vector[1]:
                        markov_stages[t][household] = 2
                    elif spins[household] < markov_vector[2]:
                        markov_stages[t][household] = 3
                    else:
                        markov_stages[t][household] = 4

                    # Alternative method: use deques to simplify the algorithm

                    # We create three sliding windows into the past, one for 12 steps, one for 24 ad one for all stages
                    # regressions_12 = cl.deque(maxlen=12)
                    # regressions_24 = cl.deque(maxlen=24)
                    # regressions_all = cl.deque()

                    # We check if during this step there was a regression
                    # if markov_stages[t][household] < markov_stages[t - 1][household]:
                    #    regression_happened = 1
                    # else:
                    #    regression_happened = 0

                    # And add the result to both queues
                    # regressions_12.append(regression_happened)
                    # regressions_24.append(regression_happened)
                    # regressions_all.append(regression_happened)

                    # Finally test the stage 5 condition per queue and a special condition when t reaches 85.
                    # if (sum(regressions_12) > 4) or \
                    #    (sum(regressions_24) > 7) or \
                    #    (sum(regressions_all) > 10) or \
                    #    t == 85:
                    #    markov_stages[t][household] = 5

                    if t >= 1:
                        # Check every timestep that occurred prior to the current timestep.
                        regressions = self.compute_regressions(markov_stages, household, 1, t)

                        if regressions > 10:
                            markov_stages[t][household] = 5
                    if t >= 12:
                        # Check the previous 12 timesteps that occurred prior to the current timestep.
                        regressions = self.compute_regressions(markov_stages, household, t - 11, t)

                        # If the number of regressive steps in the household's past 12 timesteps is greater than 4,
                        # the household transitions to stage 5.
                        if regressions > 4:
                            markov_stages[t][household] = 5
                    if t >= 24:
                        # Check the previous 24 timesteps that occurred prior to the current timestep.
                        regressions = self.compute_regressions(markov_stages, household, t - 23, t)

                        # If the number of regressive steps in the household's past 24 timesteps is greater than 7,
                        # the household transitions to stage 5.
                        if regressions > 7:
                            markov_stages[t][household] = 5
                    if t == 85:
                        # If the household has not reached stage 4 after 85 timesteps, the household transitions to
                        # stage 5.
                        if markov_stages[t][household] != 4:
                            markov_stages[t][household] = 5

        # We make a copy to be used for numerical purposes, from which drop some of the columns
        result = pd.DataFrame()
        result['guid'] = households_df['guid']
        result['huid'] = households_df['huid']
        result['Zone'] = households_df['Zone']
        result['SV'] = sv_scores
        column_names = [str(i) for i in range(1, stages + 1)]

        for i, c in enumerate(column_names):
            result[c] = markov_stages[i]

        return result

    @staticmethod
    def compute_social_vulnerability_zones(sv_result, households_df):
        """
        Compute the social vulnerability score based on dislocation attributes. Updates the dislocation dataset
        by adding a new `Zone` column and removing values with missing Zone.

        Args:
            sv_result (pd.DataFrame): output from social vulnerability analysis
            households_df (pd.DataFrame): Vector position of a household.

        Returns:
            pd.DataFrame: Social vulnerability score.

        """
        # if FIPS has 11 digits (Tract level)
        if len(sv_result["FIPS"].iloc[0]) == 11:
            households_df['blockfips'] = households_df['blockid'].apply(lambda x: str(x)[:11]).astype(str)
        # if FIPS has 12 digits (Block Group level)
        elif len(sv_result["FIPS"].iloc[0]) == 12:
            households_df['blockfips'] = households_df['blockid'].apply(lambda x: str(x)[:12]).astype(str)

        households_df = households_df.merge(sv_result[["FIPS", "zone"]], left_on="blockfips", right_on="FIPS")
        # e.g.Medium Vulnerable (zone3) extract the number 3 to construct Z3
        households_df["Zone"] = households_df["zone"].apply(lambda row: "Z"+row[-2])

        return households_df[households_df['Zone'] != 'missing']

    def compute_social_vulnerability_values(self, households_df, num_households, rng):
        """
        Compute the social vulnerability score of a household depending on its zone
        Args:
            households_df (pd.DataFrame): Information about household zones.
            num_households (int): Number of households.
            rng (np.RandomState): Random state to draw pseudo-random numbers from.
        Returns:
            pd.Series: social vulnerability scores.
        """
        # Social vulnerability zone generator: this generalizes the code in the first version
        sv_scores = np.zeros(num_households)
        zones = households_df['Zone'].to_numpy()

        for household in range(0, num_households):
            spin = rng.rand()
            zone = zones[household]

            if spin < self.__sv_generator[zone]['threshold_0']:
                sv_scores[household] = round(rng.uniform(self.__sv_generator[zone]['below_lower'],
                                                         self.__sv_generator[zone]['below_upper']), 3)

            # for zone 2, 3, 4 there is additional middle range
            elif 'threshold_1' in self.__sv_generator[zone].keys() \
                    and spin < self.__sv_generator[zone]['threshold_1']:
                sv_scores[household] = round(rng.uniform(self.__sv_generator[zone]['middle_lower'],
                                                         self.__sv_generator[zone]['middle_upper']), 3)
            else:
                sv_scores[household] = round(rng.uniform(self.__sv_generator[zone]['above_lower'],
                                                         self.__sv_generator[zone]['above_upper']), 3)

        return sv_scores

    @staticmethod
    def compute_regressions(markov_stages, household, lower, upper):
        """
        Compute regressions for a given household in the interval near t using the interval [lower, upper], and
        adjust for stage inversion at the upper boundary.

        Args:
            markov_stages (np.Array): Markov chain stages for all households.
            household (int): Index of the household.
            lower (int): Lower index to check past history.
            upper (int): Upper index to check past history.

        Returns:
            int: Number of regressions between a given time window.

        """
        regressions = 0

        # Check the previous timesteps between [lower, upper] prior to the current timestep.

        for t in range(lower, upper):
            if markov_stages[t][household] < markov_stages[t - 1][household]:
                regressions += 1

        return regressions

    def get_spec(self):
        """Get specifications of the housing serial recovery model.

        Returns:
            obj: A JSON object of specifications of the housing serial recovery model.

        """
        return {
            'name': 'housing-recovery-serial',
            'description': 'Household-level housing recovery serial model',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result CSV dataset name',
                    'type': str
                },
                {
                    'id': 't_delta',
                    'required': True,
                    'description': 'size of the analysis time step',
                    'type': float
                },
                {
                    'id': 't_final',
                    'required': True,
                    'description': 'total duration',
                    'type': float
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Seed to ensure replication of the Markov Chain path'
                                   'in connection with Population Dislocation.',
                    'type': int
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'population_dislocation_block',
                    'required': True,
                    'description': 'A csv file with population dislocation result '
                                   'aggregated to the block group level',
                    'type': ['incore:popDislocation']
                },
                {
                    'id': 'tpm',
                    'required': True,
                    'description': 'Transition probability matrix in CSV format that specifies '
                                   'the corresponding Markov chain per social vulnerability level.',
                    'type': ['incore:houseRecTransitionProbMatrix']
                },
                {
                    'id': 'initial_stage_probabilities',
                    'required': True,
                    'description': 'initial mass probability function for stage 0 of the Markov Chain',
                    'type': ['incore:houseRecInitialStageProbability']
                },
                {
                    'id': 'sv_result',
                    'required': True,
                    'description': 'A csv file with zones containing demographic factors'
                                   'qualified by a social vulnerability score',
                    'type': ['incore:socialVulnerabilityScore']
                }
            ],
            'output_datasets': [
                {
                    'id': 'ds_result',
                    'parent_type': 'housing_recovery_block',
                    'description': 'A csv file with housing recovery sequences'
                                   'at the individual household level',
                    'type': 'incore:housingRecoveryHistory'
                }
            ]
        }



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd
from pyincore import BaseAnalysis


class HousingUnitAllocation(BaseAnalysis):
    def __init__(self, incore_client):
        super(HousingUnitAllocation, self).__init__(incore_client)

    def get_spec(self):
        return {
            'name': 'housing-unit-allocation',
            'description': 'Probabilistic Housing Unit Allocation Analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result CSV dataset name',
                    'type': str
                },
                {
                    'id': 'seed',
                    'required': True,
                    'description': 'Initial seed for the probabilistic model',
                    'type': int
                },
                {
                    'id': 'iterations',
                    'required': True,
                    'description': 'No of iterations to perform the probabilistic model on',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5',
                             'ergo:buildingInventoryVer6', 'ergo:buildingInventoryVer7'],
                },
                {
                    'id': 'housing_unit_inventory',
                    'required': True,
                    'description': 'Housing Unit Inventory CSV data, aka Census Block data. Corresponds to a possible '
                                   'occupied housing unit, vacant housing unit, or a group quarters',
                    'type': ['incore:housingUnitInventory']
                },
                {
                    'id': 'address_point_inventory',
                    'required': True,
                    'description': 'CSV dataset of address locations available in a block. Corresponds to a '
                                   'specific address where a housing unit or group quarters could be assigned',
                    'type': ['incore:addressPoints']
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'description': 'A csv file with the merged dataset of the inputs, aka Probabilistic'
                                   'Housing Unit Allocation',
                    'type': 'incore:housingUnitAllocation'
                }
            ]
        }

    def run(self):
        """Merges Housing Unit Inventory, Address Point Inventory and Building Inventory.
         The results of this analysis are aggregated per structure/building. Generates
         one csv result per iteration.

        Returns:
            bool: True if successful, False otherwise

        """

        # Get seed
        seed = self.get_parameter("seed")

        # Get iterations
        iterations = self.get_parameter("iterations")

        # Get desired result name
        result_name = self.get_parameter("result_name")

        # Datasets
        bg_inv = self.get_input_dataset("buildings").get_dataframe_from_shapefile()
        pop_inv = self.get_input_dataset("housing_unit_inventory").get_dataframe_from_csv(low_memory=False)
        addr_point_inv = self.get_input_dataset("address_point_inventory").get_dataframe_from_csv(low_memory=False)

        for i in range(iterations):
            seed_i = seed + i
            hua_inventory = self.get_iteration_probabilistic_allocation(
                pop_inv, addr_point_inv, bg_inv, seed_i)
            temp_output_file = result_name + "_" + str(seed_i) + ".csv"

            # first column guid
            hua_inventory = hua_inventory[["addrptid"] + [col for col in hua_inventory.columns if col != "addrptid"]]
            # last column geometry
            hua_inventory = hua_inventory[[col for col in hua_inventory.columns if col != "geometry"] + ["geometry"]]
            self.set_result_csv_data("result",
                                     hua_inventory,
                                     temp_output_file, "dataframe")
        return True

    def prepare_housing_unit_inventory(self, housing_unit_inventory, seed):
        """Merge order to Building and Address inventories.

        Args:
            housing_unit_inventory (pd.DataFrame): Housing unit inventory.
            seed (int): Random number generator seed for reproducibility.

        Returns:
            pd.DataFrame: Sorted housing unit inventory.

        """
        size_row, size_col = housing_unit_inventory.shape

        random_generator = np.random.RandomState(seed)
        sorted_housing_unit0 = housing_unit_inventory.sort_values(by=["huid"])

        # Create Random merge order for housing unit inventory
        random_order_housing_unit = random_generator.uniform(0, 1, size_row)

        sorted_housing_unit0["randomhu"] = random_order_housing_unit

        #  gsort BlockID -LiveTypeUnit Tenure randomaorderpop
        sorted_housing_unit1 = sorted_housing_unit0.sort_values(by=["blockid", "ownershp", "vacancy", "randomhu"],
                                                                ascending=[True, True, True, True])

        # by BlockID: gen RandomMergeOrder = _n (+1 to be consistent with STATA starting from 1)
        sorted_housing_unit1["randommergeorder"] = sorted_housing_unit1.groupby(["blockid"]).cumcount() + 1

        sorted_housing_unit2 = sorted_housing_unit1.sort_values(by=["blockid", "randommergeorder"],
                                                                ascending=[True, True])
        return sorted_housing_unit2

    def merge_infrastructure_inventory(self, address_point_inventory, building_inventory):
        """Merge order to Building and Address inventories.

        Args:
            address_point_inventory (pd.DataFrame): address point inventory
            building_inventory (pd.DataFrame): building inventory

        Returns:
            pd.DataFrame: merged address and building inventories

        """
        sorted_pnt_0 = address_point_inventory.sort_values(by=["strctid"])
        sorted_bld_0 = building_inventory.sort_values(by=["strctid"])

        addresspt_building_inv = pd.merge(sorted_bld_0, sorted_pnt_0,
                                          how='outer', on="strctid",
                                          left_index=False, right_index=False,
                                          sort=True, copy=True, indicator=True,
                                          validate="1:m")

        # addresspt_building_inv = self.compare_merges(sorted_pnt_0.columns, sorted_bld_0.columns,
        #                                             addresspt_building_inv)

        match_column = set(sorted_pnt_0.columns).intersection(sorted_bld_0.columns)
        for col in match_column:
            # Compare two columns, keep one from the address, rename and drop
            if col + "_x" in addresspt_building_inv.columns and col + "_y" in addresspt_building_inv.columns:
                addresspt_building_inv[col] = addresspt_building_inv[col + "_y"]
                addresspt_building_inv = addresspt_building_inv.drop(columns=[col + "_x", col + "_y"])

        addresspt_building_inv = addresspt_building_inv.drop(columns=["_merge"])

        return addresspt_building_inv

    def prepare_infrastructure_inventory(self, seed_i: int, critical_bld_inv: pd.DataFrame):
        """Assign Random merge order to Building and Address inventories. Use main
        seed value.

        Args:
            seed_i (int): random number generator seed for reproducibility
            critical_bld_inv (pd.DataFrame): Merged inventories

        Returns:
            pd.DataFrame: Sorted merged critical infrastructure

        """
        size_row, size_col = critical_bld_inv.shape

        sort_critical_bld_0 = critical_bld_inv.sort_values(by=["addrptid"])

        seed_i2: int = seed_i + 1
        random_generator = np.random.RandomState(seed_i2)

        randomap = random_generator.uniform(0, 1, size_row)
        sort_critical_bld_0["randomap"] = randomap
        sort_critical_bld_1 = sort_critical_bld_0.sort_values(by=["blockid", "residential", "huestimate", "randomap"],
                                                              ascending=[True, False, True, True])

        # +1 to be consistent with STATA starting from 1
        sort_critical_bld_1["randommergeorder"] = sort_critical_bld_1.groupby(["blockid"]).cumcount() + 1

        sort_critical_bld_2 = sort_critical_bld_1.sort_values(by=["blockid", "randommergeorder"],
                                                              ascending=[True, False])
        return sort_critical_bld_2

    def merge_inventories(self, sorted_housing_unit: pd.DataFrame, sorted_infrastructure: pd.DataFrame):
        """Merge (Sorted) Housing Unit Inventory and (Sorted) Infrastructure Inventory.

        Args:
            sorted_housing_unit (pd.DataFrame):  Sorted Housing Unit Inventory
            sorted_infrastructure (pd.DataFrame):  Sorted infrastructure inventory. This includes
                Building inventory and Address point inventory.

        Returns:
            pd.DataFrame: Final merge of all four inventories

        """
        huap_inventory = pd.merge(sorted_infrastructure, sorted_housing_unit,
                                  how='outer', left_on=["blockid", "randommergeorder"],
                                  right_on=["blockid", "randommergeorder"],
                                  sort=True, suffixes=("_x", "_y"),
                                  copy=True, indicator=True)
        huap_inventory = huap_inventory.rename(columns={"_merge": "aphumerge"})
        # check for duplicate columns from merge
        huap_inventory = self.compare_merges(sorted_housing_unit.columns,
                                             sorted_infrastructure.columns,
                                             huap_inventory)

        output = huap_inventory.sort_values(by=["aphumerge", "blockid"], ascending=[False, True])

        return output

    def get_iteration_probabilistic_allocation(self, housing_unit_inventory, address_point_inventory,
                                               building_inventory, seed):
        """Merge inventories

        Args:
            housing_unit_inventory (pd.DataFrame): Housing Unit Inventory
            address_point_inventory (pd.DataFrame): Address Point inventory
            building_inventory (pd.DataFrame): Building inventory
            seed (int): random number generator seed for reproducibility

            Returns:
                pd.DataFrame: Merged table

        """
        sorted_housing_unit = self.prepare_housing_unit_inventory(housing_unit_inventory, seed)

        critical_building_inv = self.merge_infrastructure_inventory(address_point_inventory, building_inventory)
        sorted_infrastructure = self.prepare_infrastructure_inventory(seed, critical_building_inv)

        output = self.merge_inventories(sorted_housing_unit, sorted_infrastructure)
        return output

    # utility functions
    def compare_merges(self, table1_cols, table2_cols, table_merged):
        """Compare two lists of columns and run compare columns on columns in both lists.
        It assumes that suffixes are _x and _y

        Args:
            table1_cols (Index):         columns in table 1
            table2_cols (Index):         columns in table 2
            table_merged (pd.DataFrame):merged table

            Returns:
                pd.DataFrame: Merged table

        """
        match_column = set(table1_cols).intersection(table2_cols)
        for col in match_column:
            # Compare two columns and marked similarity or rename and drop
            if col + "_x" in table_merged.columns and col + "_y" in table_merged.columns:
                table_merged = self.compare_columns(table_merged, col + "_x", col + "_y", True)

        return table_merged

    def compare_columns(self, table, col1, col2, drop):
        """Compare two columns. If not equal create Tru/False column, if equal rename one of them
        with the base name and drop the other.

        Args:
            table (pd.DataFrame):      Data Frame table
            col1 (pd.Series):          Column 1
            col2 (pd.Series):          Column 2
            drop (bool):               rename and drop column

        Returns:
            pd.DataFrame: Table with True/False column

        """
        # Values in columns match or not, add True/False column
        table.loc[table[col1] == table[col2], col1 + "-" + col2] = True
        table.loc[table[col1] != table[col2], col1 + "-" + col2] = False

        if table[col1].equals(table[col2]):
            col1_base = col1.rsplit("_", 1)[0]
            col2_base = col1.rsplit("_", 1)[0]
            if col1_base == col2_base and drop:
                table[col1_base] = table[col1]
                table = table.drop(columns=[col1, col2, col1 + "-" + col2])

        return table



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore.analyses.saltlakecge.equationlib import *
import pandas as pd
import numpy as np


# Return initial values function
def baseValue(vars, soln, eqName):
    if vars.getInfo(eqName)['size'] == 1:
        basevalues = vars.initialVals[vars.getIndex(eqName)]

    else:
        rows = vars.getInfo(eqName)['rows']
        rws = []
        if vars.getInfo(eqName)['size'] == vars.getInfo(eqName)['nrows']:
            for rr in rows:
                number = vars.initialVals[vars.getIndex(eqName, row=rr)]
                rws.append(number)
            basevalues = pd.DataFrame(rws, index=rows)

        else:
            cols = vars.getInfo(eqName)['cols']
            for rr in rows:
                clms = []
                for cc in cols:
                    number = vars.initialVals[vars.getIndex(eqName, row=rr, col=cc)]
                    clms.append(number)
                rws.append(clms)
            basevalues = pd.DataFrame(rws, index=rows, columns=cols)

    return basevalues


# Get the new values function
def newValue(vars, soln, eqName, ittr):
    #    for scalar
    if vars.getInfo(eqName)['size'] == 1:
        newvalue = soln[ittr][vars.getIndex(eqName)]

    else:
        rows = vars.getInfo(eqName)['rows']
        rws = []
        #       vectors
        if vars.getInfo(eqName)['size'] == vars.getInfo(eqName)['nrows']:
            for rr in rows:
                number = soln[ittr][vars.getIndex(eqName, row=rr)]
                rws.append(number)
            newvalue = pd.DataFrame(rws, index=rows)
        #       data frames
        else:
            for rr in rows:
                cols = vars.getInfo(eqName)['cols']
                clms = []
                for cc in cols:
                    number = soln[ittr][vars.getIndex(eqName, row=rr, col=cc)]
                    clms.append(number)
                rws.append(clms)
            newvalue = pd.DataFrame(rws, index=rows, columns=cols)

    return newvalue


# Get differences function
def getDiff(vars, soln, eqName, ittr):
    difference = newValue(vars, soln, eqName, ittr) - baseValue(vars, soln, eqName)
    return difference



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import networkx as nx


class InfrastructureNetwork(object):
    """
    Stores information of the infrastructure network

    Attributes
    ----------
    G : networkx.DiGraph
        The networkx graph object that stores node, arc, and interdependency information
    S : list
        List of geographical spaces on which the network lays
    id : int
        Id of the network
    """

    def __init__(self, id):
        self.G = nx.DiGraph()
        self.S = []
        self.id = id

    def copy(self):
        """
        This function copies the current :class:`InfrastructureNetwork` object

        Returns
        -------
        new_net: :class:`InfrastructureNetwork`
            Copy of the current infrastructure network object
        """
        new_net = InfrastructureNetwork(self.id)
        new_net.G = self.G.copy()
        new_net.S = [s for s in self.S]
        return new_net

    def update_with_strategy(self, player_strategy):
        """
        This function modify the functionality of node and arc per a given strategy

        Parameters
        ----------
        player_strategy : list
            Given strategy, where the first list item shows the functionality of nodes, and the second one is for arcs

        Returns
        -------
        None.

        """
        for q in player_strategy[0]:
            node = q
            strat = player_strategy[0][q]
            self.G.node[q]['data']['inf_data'].repaired = round(strat['repair'])
            self.G.node[q]['data']['inf_data'].functionality = round(strat['w'])
        for q in player_strategy[1]:
            src = q[0]
            dst = q[1]
            strat = player_strategy[1][q]
            self.G[src][dst]['data']['inf_data'].repaired = round(strat['repair'])
            self.G[src][dst]['data']['inf_data'].functionality = round(strat['y'])

    def get_clusters(self, layer):
        """
        This function find the clusters in a layer of the network

        Parameters
        ----------
        layer : int
            The id of the desired layer

        Returns
        -------
            : list
            List of layer components

        """
        g_prime_nodes = [n[0] for n in self.G.nodes(data=True) if
                         n[1]['data']['inf_data'].net_id == layer and n[1]['data']['inf_data'].functionality == 1.0]
        g_prime = nx.DiGraph(self.G.subgraph(g_prime_nodes).copy())
        g_prime.remove_edges_from(
            [(u, v) for u, v, a in g_prime.edges(data=True) if a['data']['inf_data'].functionality == 0.0])
        # print nx.connected_components(g_prime.to_undirected())
        return list(nx.connected_components(g_prime.to_undirected()))

    def gc_size(self, layer):
        """
        This function finds the size of the largest component in a layer of the network

        Parameters
        ----------
        layer : int
            The id of the desired layer

        Returns
        -------
            : int
            Size of the largest component in the layer
        """
        g_prime_nodes = [n[0] for n in self.G.nodes(data=True) if
                         n[1]['data']['inf_data'].net_id == layer and n[1]['data']['inf_data'].functionality == 1.0]
        g_prime = nx.Graph(self.G.subgraph(g_prime_nodes))
        g_prime.remove_edges_from(
            [(u, v) for u, v, a in g_prime.edges(data=True) if a['data']['inf_data'].functionality == 0.0])
        cc = nx.connected_components(g_prime.to_undirected())
        if cc:
            # if len(list(cc)) == 1:
            #    print "I'm here"
            #    return len(list(cc)[0])
            # cc_list=list(cc)
            # print "Length",len(cc_list)
            # if len(cc_list) == 1:
            #    return len(cc_list[0])
            return len(max(cc, key=len))
        else:
            return 0

    def to_game_file(self, layers=None):
        """
        This function writes the multi-defender security games.

        Parameters
        ----------
        layers : list
            List of layers in the game.

        Returns
        -------
        None.

        """
        if layers is None:
            layers = [1, 3]
        with open("../results/indp_gamefile.game") as f:
            num_players = len(layers)
            num_targets = len([n for n in self.G.nodes(data=True) if n[1]['data']['inf_data'].net_id in layers])
            f.write(str(num_players) + "," + str(num_targets) + ",2\n")
            for layer in range(len(layers)):
                layer_nodes = [n for n in self.G.nodes(data=True) if n[1]['data']['inf_data'].net_id == layers[layer]]
                for node in layer_nodes:
                    def_values = [0.0] * len(layers)
                    def_values[layer] = abs(node[1]['data']['inf_data'].demand)
                    atk_values = sum(def_values)
                    f.write(str(layer) + "\n")
                    # f.write("0.0,1.0")
                    for v in def_values:
                        f.write(str(v) + "\n")
                    f.write(str(atk_values) + "\n")

    def to_csv(self, filename="infrastructure_adj.csv"):
        """
        This function writes the object to a csv file

        Parameters
        ----------
        filename : str
            Name of the file to which the network should be written

        Returns
        -------
        None.

        """
        with open(filename, 'w') as f:
            for u, v, a in self.G.edges(data=True):
                f.write(str(u[0]) + "." + str(u[1]) + "," + str(v[0]) + "." + str(v[1]) + "\n")



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from pyincore import BaseAnalysis, AnalysisUtil
from itertools import repeat
from typing import List
import concurrent.futures
import collections


class MeanDamage(BaseAnalysis):
    """
    Args:
        incore_client (IncoreClient): Service authentication.
    """

    def __init__(self, incore_client):
        super(MeanDamage, self).__init__(incore_client)

    def get_spec(self):
        """Get specifications of the mean damage calculation.

        Returns:
            obj: A JSON object of specifications of the mean damage calculation.

        """
        return {
            'name': 'mean-damage',
            'description': 'calculate the mean and expected damage using damage ratio table',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'damage_interval_keys',
                    'required': True,
                    'description': 'Column name of the damage interval must be four and ranged in order',
                    'type': List[str]
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'damage',
                    'required': True,
                    'description': 'damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer4',
                             'ergo:buildingDamageVer5',
                             'ergo:buildingDamageVer6',
                             'ergo:nsBuildingInventoryDamage',
                             'ergo:nsBuildingInventoryDamageVer2',
                             'ergo:nsBuildingInventoryDamageVer3',
                             'ergo:bridgeDamage',
                             'ergo:bridgeDamageVer2',
                             'ergo:bridgeDamageVer3',
                             'ergo:roadDamage',
                             'ergo:roadDamageVer2',
                             'ergo:roadDamageVer3']
                },
                {
                    'id': 'dmg_ratios',
                    'required': True,
                    'description': 'Damage Ratios table',
                    'type': ['ergo:buildingDamageRatios',
                             'ergo:bridgeDamageRatios',
                             'ergo:buildingContentDamageRatios',
                             'ergo:buildingASDamageRatios',
                             'ergo:buildingDSDamageRatios',
                             'ergo:roadDamageRatios']
                },

            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'description': 'CSV file of mean damage',
                    'type': 'ergo:meanDamage'
                }
            ]
        }

    def run(self):
        """Executes mean damage calculation."""

        # read in file and parameters
        damage = self.get_input_dataset("damage").get_csv_reader()
        damage_result = AnalysisUtil.get_csv_table_rows(damage, ignore_first_row=False)

        dmg_ratio_csv = self.get_input_dataset("dmg_ratios").get_csv_reader()
        dmg_ratio_tbl = AnalysisUtil.get_csv_table_rows(dmg_ratio_csv)

        # setting number of cpus to use
        user_defined_cpu = 1
        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(
            damage_result), user_defined_cpu)

        avg_bulk_input_size = int(len(damage_result) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = damage_result
        while count < len(inventory_list):
            inventory_args.append(
                inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        results = self.mean_damage_concurrent_future(
            self.mean_damage_bulk_input, num_workers,
            inventory_args, repeat(dmg_ratio_tbl))
        self.set_result_csv_data("result", results,
                                 name=self.get_parameter("result_name"))
        return True

    def mean_damage_concurrent_future(self, function_name, parallelism, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            parallelism (int): Number of workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with building damage values and other data/metadata.

        """
        output = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=parallelism) as executor:
            for ret in executor.map(function_name, *args):
                output.extend(ret)

        return output

    def mean_damage_bulk_input(self, damage, dmg_ratio_tbl):
        """Run analysis for mean damage calculation

        Args:
            damage (obj): output of building/bridge/waterfacility/epn damage that has damage interval
            dmg_ratio_tbl (list): damage ratio table

        Returns:
            list: A list of ordered dictionaries with mean damage, deviation, and other data/metadata.

        """
        damage_interval_keys = self.get_parameter("damage_interval_keys")

        data_type = self.get_input_dataset("damage").data_type

        if ":bridgeDamage" in data_type:
            is_bridge = True
        else:
            is_bridge = False

        result = []
        for dmg in damage:
            result.append(
                self.mean_damage(dmg, dmg_ratio_tbl, damage_interval_keys,
                                 is_bridge))

        return result

    def mean_damage(self, dmg, dmg_ratio_tbl, damage_interval_keys, is_bridge):
        """Calculates mean damage based on damage probabilities and damage ratios

        Args:
            dmg (obj): dmg analysis output for a single entity in the built environment
            dmg_ratio_tbl (list): dmg ratio table.
            damage_interval_keys (list): damage interval keys
            is_bridge (bool): a boolean to indicate if the inventory type is bridge.
            Bridge has its own way of calculating mean damage

        Returns:
            OrderedDict: A dictionary with mean damage, deviation, and other data/metadata.

        """
        results = collections.OrderedDict()
        results.update(dmg)

        if is_bridge:
            # need to calculate bridge span
            if "spans" in dmg.keys() and dmg['spans'] is not None \
                    and dmg["spans"].isdigit():
                bridge_spans = int(dmg["spans"])
            else:
                bridge_spans = 1

            if bridge_spans > 10:
                bridge_spans = 10
                print("A bridge was found with greater than 10 spans: "
                      + dmg['guid'] + ". Default to 10 bridge spans.")

            mean_damage = AnalysisUtil.calculate_mean_damage(dmg_ratio_tbl,
                                                             dmg,
                                                             damage_interval_keys,
                                                             is_bridge,
                                                             bridge_spans)
        else:
            mean_damage = AnalysisUtil.calculate_mean_damage(dmg_ratio_tbl,
                                                             dmg,
                                                             damage_interval_keys,
                                                             is_bridge)
        results.update(mean_damage)

        # bridge doesn't calculates deviation
        if not is_bridge:
            mean_damage_dev = AnalysisUtil.calculate_mean_damage_std_deviation(
                dmg_ratio_tbl, dmg, mean_damage['meandamage'],
                damage_interval_keys)
            results.update(mean_damage_dev)
        else:
            expected_damage = AnalysisUtil.get_expected_damage(
                mean_damage['meandamage'],
                dmg_ratio_tbl)
            results['expectval'] = expected_damage

        return results



# Copyright (c) 2020 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import json
import math

import numpy

from pyincore import globals as pyglobals
from pyincore.utils import evaluateexpression

logger = pyglobals.LOGGER


class DFR3Curve:
    """A class to represent a DFR3 curve."""

    def __init__(self, curve_parameters):
        self.rules = curve_parameters['rules']
        self.return_type = curve_parameters['returnType']

        for rule in self.rules:
            rule["expression"] = rule["expression"].replace("^", "**")
        self.description = curve_parameters['description']

    def solve_curve_expression(self, hazard_values: dict, curve_parameters: dict, **kwargs):
        """Evaluates expression of the curve.

        Args:
            hazard_values (dict): Hazard values. Only applicable to fragilities
            curve_parameters (dict): Curve parameters.
            **kwargs: Keyword arguments.

        Returns:
            any: Result of the evaluated expression. Can be float, numpy.ndarray etc.

        """
        parameters = {}
        mapped_demand_types = {}
        # For all curve parameters:
        # 1. Figure out if parameter name needs to be mapped (i.e. the name contains forbidden characters)
        # 2. Fetch all parameters listed in the curve from kwargs and if there are not in kwargs, use default values
        # from the curve.
        for parameter in curve_parameters:
            # if default exists, use default
            if "expression" in parameter and parameter["expression"] is not None:
                parameters[parameter["name"]] = evaluateexpression.evaluate(parameter["expression"], parameters)
            else:
                parameters[parameter["name"]] = None

            # e.g. map point_two_sec_sa to its full name (0.2 Sec Sa)
            if "fullName" in parameter and parameter["fullName"] is not None:
                mapped_demand_types[parameter["fullName"]] = parameter["name"]

            # else overwrite with real values; make sure it handles case sensitivity
            for kwargs_key, kwargs_value in kwargs.items():
                if "fullName" in parameter and parameter["fullName"] is not None:
                    if parameter["fullName"].lower() == kwargs_key.lower():
                        parameters[parameter["name"]] = kwargs_value
                elif parameter["name"].lower() == kwargs_key.lower():
                    parameters[parameter["name"]] = kwargs_value

        # use hazard values if present
        # consider case insensitive situation
        for key, value in hazard_values.items():
            if key in mapped_demand_types:
                key = mapped_demand_types[key]
            for parameter_key in parameters.keys():
                if parameter_key.lower() == key.lower():
                    if value is not None:
                        parameters[parameter_key] = value
                    else:
                        # returning 0 even if a single demand value is None, assumes there is no hazard exposure. TBD
                        return 0.0

        eval_result = None
        for rule in self.rules:
            if "condition" not in rule or rule["condition"] is None:
                eval_result = evaluateexpression.evaluate(rule["expression"], parameters)
            else:
                conditions_met = []
                for condition in rule["condition"]:
                    if evaluateexpression.evaluate(condition, parameters):
                        conditions_met.append(True)
                    else:
                        conditions_met.append(False)
                        break
                if all(conditions_met):
                    eval_result = evaluateexpression.evaluate(rule["expression"], parameters)
                    break

        if isinstance(eval_result, numpy.ndarray) or isinstance(eval_result, list):  # for repair curves etc.
            return eval_result
        else:  # for fragility curves the return is a float
            if eval_result is None:
                eval_result = 0.0
            if math.isnan(eval_result):
                error_msg = "Unable to calculate limit state."
                if self.rules:
                    error_msg += " Evaluation failed for expression: \n" + json.dumps(self.rules) + "\n"
                    error_msg += "Provided Inputs: \n" + json.dumps(hazard_values) + "\n" + json.dumps(kwargs)

                raise ValueError(error_msg)

            return eval_result

    def solve_curve_for_inverse(self, hazard_values: dict, curve_parameters: dict, **kwargs):
        """Evaluates expression of the curve by calculating its inverse. Example, ppf for cdf. Only supports cdf() for
         now. More inverse methods may be added in the future.

        Args:
            hazard_values (dict): Hazard values. Only applicable to fragilities
            curve_parameters (dict): Curve parameters.
            **kwargs: Keyword arguments.

        Returns:
            any: Result of the evaluated inverse expression. Can be float, numpy.ndarray etc.

        """
        inverse_rules = []
        actual_rules = self.rules
        for rule in self.rules:
            if ".cdf(" in rule['expression']:
                new_exp = rule['expression'].replace(".cdf(", ".ppf(")
                inverse_rules.append({'condition': rule['condition'], 'expression': new_exp})
            else:
                raise KeyError("Inverse does not exist for the provided expression. exiting..")

        self.rules = inverse_rules
        inverse = self.solve_curve_expression(hazard_values=hazard_values, curve_parameters=curve_parameters, **kwargs)
        self.rules = actual_rules  # swap the original rules back so further calculations are not affected
        return inverse

    def get_building_period(self, curve_parameters, **kwargs):
        """
                Get building period from the fragility curve.

        Args:
            curve_parameters (dict): Fragility curve parameters.
            **kwargs: Keyword arguments.

        Returns:
            float: Building period.

        """

        period = 0.0
        num_stories = 1.0
        for parameter in curve_parameters:
            # if default exists, use default
            if parameter["name"] == "num_stories" and "expression" in parameter and parameter["expression"] is not None:
                num_stories = evaluateexpression.evaluate(parameter["expression"])

            # if exist in building inventory
            for kwargs_key, kwargs_value in kwargs.items():
                if kwargs_key.lower() == "num_stories" and kwargs_value is not None and kwargs_value > 0:
                    num_stories = kwargs_value

            # calculate period
            if parameter["name"] == "period" and "expression" in parameter and parameter["expression"] is not None:
                period = evaluateexpression.evaluate(parameter["expression"], {"num_stories": num_stories})

        return period



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


"""
Electric Power Facility Restoration
"""

import numpy as np

from typing import List
from pyincore import BaseAnalysis, RestorationService, AnalysisUtil
from pyincore.models.restorationcurveset import RestorationCurveSet


class EpfRestoration(BaseAnalysis):
    """Computes electric power facility restoration for an earthquake exposure."""

    def __init__(self, incore_client):
        self.restorationsvc = RestorationService(incore_client)
        super(EpfRestoration, self).__init__(incore_client)

    def run(self):
        """Performs Water facility restoration analysis by using the parameters from the spec
        and creates an output dataset in csv format

        Returns:
            bool: True if successful, False otherwise

        """
        inventory_list = list(self.get_input_dataset("epfs").get_inventory_reader())
        mapping_set = self.get_input_dataset("dfr3_mapping_set")

        restoration_key = self.get_parameter("restoration_key")
        if restoration_key is None:
            restoration_key = "Restoration ID Code"

        end_time = self.get_parameter("end_time")
        if end_time is None:
            end_time = 365.0

        time_interval = self.get_parameter("time_interval")
        if time_interval is None:
            time_interval = 1

        pf_interval = self.get_parameter("pf_interval")
        if pf_interval is None:
            pf_interval = 0.1

        discretized_days = self.get_parameter("discretized_days")
        if discretized_days is None:
            discretized_days = [1, 3, 7, 30, 90]

        damage = self.get_input_dataset("damage").get_csv_reader()
        damage_result = AnalysisUtil.get_csv_table_rows(damage, ignore_first_row=False)

        (inventory_restoration_map, pf_results, time_results, func_results,
         repair_times) = self.electricpowerfacility_restoration(
            inventory_list, damage_result, mapping_set, restoration_key, end_time, time_interval, pf_interval, 
            discretized_days)

        self.set_result_csv_data("inventory_restoration_map", inventory_restoration_map,
                                 name="inventory_restoration_map_" + self.get_parameter("result_name"))
        self.set_result_csv_data("pf_results", time_results, name="percentage_of_functionality_" +
                                                                  self.get_parameter("result_name"))
        self.set_result_csv_data("time_results", pf_results, name="reptime_" + self.get_parameter("result_name"))
        self.set_result_csv_data("func_results", func_results,
                                 name=self.get_parameter("result_name") + "_discretized_restoration")
        self.set_result_csv_data("repair_times", repair_times, name="full_reptime_" + self.get_parameter(
            "result_name"))

        return True

    def electricpowerfacility_restoration(self, inventory_list, damage_result, mapping_set, restoration_key, end_time,
                                          time_interval, pf_interval, discretized_days):
        """Gets applicable restoration curve set and calculates restoration time and functionality

        Args:
            inventory_list (list): Multiple EPF facilities from input inventory set.
            damage_result: EPF damage
            mapping_set (class): Restoration Mapping Set
            restoration_key (str): Restoration Key to determine which curve to use. E.g. Restoration ID Code
            end_time (float): User specified end repair time
            time_interval (float): Increment interval of repair time. Default to 1 (1 day)
            pf_interval (float): Increment interval of percentage of functionality. Default 0.1 (10%)
            discretized_days (list): Days to compute discretized restoration (e.g. 1, 3, 7, 30, 90)

        Returns:
            time_results (list): Given Percentage of functionality, the change of repair time
            pf_results (list): Given Repair time, change of the percentage of functionality

        """
        # Obtain the restoration id for each electric facilities
        inventory_restoration_map = []

        # Obtain classification for each electric facility, used to lookup discretized functionality
        inventory_class_map = {}
        restoration_sets = self.restorationsvc.match_inventory(
            self.get_input_dataset("dfr3_mapping_set"), inventory_list, restoration_key)
        for inventory in inventory_list:
            if inventory["id"] in restoration_sets.keys():
                restoration_set_id = restoration_sets[inventory["id"]].id
            else:
                restoration_set_id = None

            inventory_restoration_map.append({"guid": inventory['properties']['guid'],
                                              "restoration_id": restoration_set_id})

            restoration_curve_set = restoration_sets[inventory["id"]]

            # For each facility, get the discretized restoration from the continuous curve
            discretized_restoration = AnalysisUtil.get_discretized_restoration(restoration_curve_set, discretized_days)
            inventory_class_map[inventory['properties']['guid']] = discretized_restoration

        time_results = []
        pf_results = []
        repair_time = {}

        for mapping in mapping_set.mappings:
            # get restoration curves
            # if it's string:id; then need to fetch it from remote and cast to restorationcurveset object
            restoration_curve_set = mapping.entry[restoration_key]
            if isinstance(restoration_curve_set, str):
                restoration_curve_set = RestorationCurveSet(self.restorationsvc.get_dfr3_set(restoration_curve_set))

            # given time calculate pf
            time = np.arange(0, end_time + time_interval, time_interval)
            for t in time:
                pf_results.append({
                    "restoration_id": restoration_curve_set.id,
                    "time": t,
                    **restoration_curve_set.calculate_restoration_rates(time=t)
                })

            # given pf calculate time
            pf = np.arange(0, 1 + pf_interval, pf_interval)
            for p in pf:
                new_dict = {}
                t_res = restoration_curve_set.calculate_inverse_restoration_rates(time=p)
                for key, value in t_res.items():
                    new_dict.update({"time_" + key: value})
                time_results.append({
                    "restoration_id": restoration_curve_set.id,
                    "percentage_of_functionality": p,
                    **new_dict
                })

            repair_time[restoration_curve_set.id] = restoration_curve_set.calculate_inverse_restoration_rates(time=0.99)

        # Compute discretized restoration
        func_result = []
        for dmg in damage_result:
            guid = dmg['guid']
            
            # Dictionary of discretized restoration functionality
            rest_dict = inventory_class_map[guid]

            ds_0, ds_1, ds_2, ds_3, ds_4 = dmg['DS_0'], dmg['DS_1'], dmg['DS_2'], dmg['DS_3'], dmg['DS_4']
            result_dict = {}
            for time in discretized_days:
                key = "day" + str(time)
                # Only compute if we have damage
                if ds_0:
                    functionality = (rest_dict[key][0] * float(ds_0) + rest_dict[key][1] * float(ds_1) + rest_dict[
                        key][2] * float(ds_2) + rest_dict[key][3] * float(ds_3) + rest_dict[key][4] * float(ds_4))
                    result_dict.update({str(key): functionality})

            func_result.append({"guid": guid, **result_dict})

        repair_times = []
        for inventory in inventory_restoration_map:
            repair_times.append({"guid": inventory["guid"], **repair_time[inventory["restoration_id"]]})

        return inventory_restoration_map, pf_results, time_results, func_result, repair_times

    def get_spec(self):
        return {
            'name': 'electric-power-facility-restoration',
            'description': 'electric power facility restoration analysis',
            'input_parameters': [
                {
                    'id': 'restoration_key',
                    'required': False,
                    'description': 'restoration key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'end_time',
                    'required': False,
                    'description': 'end time. Default to 365.',
                    'type': float
                },
                {
                    'id': 'time_interval',
                    'required': False,
                    'description': 'incremental interval for time in days. Default to 1',
                    'type': float
                },
                {
                    'id': 'pf_interval',
                    'required': False,
                    'description': 'incremental interval for percentage of functionality. Default to 0.1',
                    'type': float
                },
                {
                    'id': 'discretized_days',
                    'required': False,
                    'description': 'Discretized days to compute functionality',
                    'type': List[int]
                }
            ],
            'input_datasets': [
                {
                    'id': 'epfs',
                    'required': True,
                    'description': 'Electric Power Facility Inventory',
                    'type': ['incore:epf', 'ergo:epf'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                },
                {

                    'id': 'damage',
                    'required': True,
                    'description': 'damage result that has damage intervals in it',
                    'type': [
                             'incore:epfDamage',
                             'incore:epfDamageVer2',
                             'incore:epfDamageVer3']
                }



            ],
            'output_datasets': [
                {
                    'id': "inventory_restoration_map",
                    'parent_type': '',
                    'description': 'A csv file recording the mapping relationship between GUID and restoration id '
                                   'applicable.',
                    'type': 'incore:inventoryRestorationMap'
                },
                {
                    'id': 'pf_results',
                    'parent_type': '',
                    'description': 'A csv file recording functionality change with time for each class and limit '
                                   'state.',
                    'type': 'incore:epfRestorationFunc'
                },
                {
                    'id': 'time_results',
                    'parent_type': '',
                    'description': 'A csv file recording repair time at certain functionality recovery for each class '
                                   'and limit state.',
                    'type': 'incore:epfRestorationTime'
                },
                {

                    'id': 'func_results',
                    'parent_type': '',
                    'description': 'A csv file recording discretized functionality over time',
                    'type': 'incore:epfDiscretizedRestorationFunc'
                },
                {
                    'id': 'repair_times',
                    'parent_type': '',
                    'description': 'A csv file recording repair time at full functionality recovery for each guid '
                                   'and limit state.',
                    'type': 'incore:epfRepairTime'
                }

            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import concurrent.futures

from pyincore import AnalysisUtil
from pyincore import BaseAnalysis


class PipelineRepairCost(BaseAnalysis):
    """Computes pipeline repair cost.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(PipelineRepairCost, self).__init__(incore_client)

    def run(self):
        """Executes pipline facility repair cost analysis."""

        pipeline_df = self.get_input_dataset("pipeline").get_dataframe_from_shapefile()
        pipeline_dmg_df = self.get_input_dataset("pipeline_dmg").get_dataframe_from_csv()
        replacement_cost = self.get_input_dataset("replacement_cost").get_dataframe_from_csv()

        # join damage, replacement cost, with original inventory
        pipeline_df = pipeline_df.merge(pipeline_dmg_df, on="guid")
        pipeline_df = pipeline_df.merge(replacement_cost, on="guid")
        pipeline_set = pipeline_df.to_dict(orient="records")

        user_defined_cpu = 1
        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(pipeline_set), user_defined_cpu)

        avg_bulk_input_size = int(len(pipeline_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(pipeline_set)
        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        repair_costs = self.pipeline_repair_cost_concurrent_future(self.pipeline_repair_cost_bulk_input, num_workers,
                                                                   inventory_args)
        self.set_result_csv_data("result", repair_costs, name=self.get_parameter("result_name") + "_repair_cost")

        return True

    def pipeline_repair_cost_concurrent_future(self, function_name, num_workers, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with pipeline repair cost values and other data/metadata.

        """

        output = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1 in executor.map(function_name, *args):
                output.extend(ret1)

        return output

    def pipeline_repair_cost_bulk_input(self, pipelines):
        """Run analysis for multiple pipelines.

        Args:
            pipelines (list): Multiple pipelines from input inventory set.

        Returns:
            list: A list of ordered dictionaries with pipeline repair cost values and other data/metadata.

        """
        # read in the damage ratio tables
        pipeline_dmg_ratios_csv = self.get_input_dataset("pipeline_dmg_ratios").get_csv_reader()
        dmg_ratio_tbl = AnalysisUtil.get_csv_table_rows(pipeline_dmg_ratios_csv, ignore_first_row=False)

        segment_length = self.get_parameter("segment_length")
        if segment_length is None:
            segment_length = 20  # 20 feet

        diameter = self.get_parameter("diameter")
        if diameter is None:
            diameter = 20  # 20 inch

        repair_costs = []

        for pipeline in pipelines:
            pipe_length = pipeline["length"]  # kilometer
            pipe_length_ft = pipeline["length"] * 3280.84  # foot

            rc = dict()
            rc["guid"] = pipeline["guid"]
            repair_cost = 0

            # read in damage ratio for break and leak
            dr_break = 0
            dr_leak = 0
            if pipeline["diameter"] > diameter:
                for dmg_ratio_row in dmg_ratio_tbl:
                    if dmg_ratio_row["Inventory Type"] == ">" + str(diameter) + " in" and \
                            dmg_ratio_row["Damage State"] == "break":
                        dr_break = float(dmg_ratio_row["Best Mean Damage Ratio"])
                    if dmg_ratio_row["Inventory Type"] == ">" + str(diameter) + " in" and \
                            dmg_ratio_row["Damage State"] == "leak":
                        dr_leak = float(dmg_ratio_row["Best Mean Damage Ratio"])
            else:
                for dmg_ratio_row in dmg_ratio_tbl:
                    if dmg_ratio_row["Inventory Type"] == "<" + str(diameter) + " in" and \
                            dmg_ratio_row["Damage State"] == "break":
                        dr_break = float(dmg_ratio_row["Best Mean Damage Ratio"])
                    if dmg_ratio_row["Inventory Type"] == "<" + str(diameter) + " in" and \
                            dmg_ratio_row["Damage State"] == "leak":
                        dr_leak = float(dmg_ratio_row["Best Mean Damage Ratio"])

            num_segment = pipe_length_ft / segment_length

            num_breaks = pipeline["breakrate"] * pipe_length
            if num_breaks > num_segment:
                repair_cost += pipeline["replacement_cost"] * dr_break
            else:
                repair_cost += pipeline["replacement_cost"] / num_segment * num_breaks * dr_break

            num_leaks = pipeline["leakrate"] * pipe_length
            if num_leaks > num_segment:
                repair_cost += pipeline["replacement_cost"] * dr_leak
            else:
                repair_cost += pipeline["replacement_cost"] / num_segment * num_leaks * dr_leak
            repair_cost = min(repair_cost, pipeline["replacement_cost"])

            rc["budget"] = repair_cost
            rc["repaircost"] = repair_cost

            repair_costs.append(rc)

        return repair_costs

    def get_spec(self):
        """Get specifications of the pipeline repair cost analysis.

        Returns:
            obj: A JSON object of specifications of the pipeline repair cost analysis.

        """
        return {
            "name": "pipeline-repair-cost",
            "description": "Pipeline repair cost analysis.",
            "input_parameters": [
                {
                    "id": "result_name",
                    "required": True,
                    "description": "A name of the resulting dataset",
                    "type": str
                },
                {
                    "id": "num_cpu",
                    "required": False,
                    "description": "If using parallel execution, the number of cpus to request.",
                    "type": int
                },
                {
                    "id": "diameter",
                    "required": False,
                    "description": "Pipeline diameter cutoff assumption for different damage ratios. Default is 20 "
                                   "inches",
                    "type": int
                },
                {
                    "id": "segment_length",
                    "required": False,
                    "description": "Segment length assumption. Default is 20 feet",
                    "type": int
                },
            ],
            "input_datasets": [
                {
                    "id": "pipeline",
                    "required": True,
                    "description": "Pipeline Inventory",
                    "type": ["ergo:buriedPipelineTopology", "ergo:pipeline"],
                },
                {
                    "id": "replacement_cost",
                    "required": True,
                    "description": "Repair cost of the node in the complete damage state (= Replacement cost)",
                    "type": ["incore:replacementCost"],
                },
                {
                    "id": "pipeline_dmg",
                    "required": True,
                    "description": "pipeline damage from PipelineDamageRepairRate Analysis",
                    "type": ["ergo:pipelineDamageVer3"]
                },
                {
                    "id": "pipeline_dmg_ratios",
                    "required": True,
                    "description": "Damage Ratios table",
                    "type": ["incore:pipelineDamageRatios"]
                },
            ],
            "output_datasets": [
                {
                    "id": "result",
                    "parent_type": "pipelines",
                    "description": "A csv file with repair cost for each pipeline",
                    "type": "incore:pipelineRepairCost"
                }
            ]
        }



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


from pyincore.analyses.transportationrecovery.nsga2 import Solution
from pyincore.analyses.transportationrecovery.nsga2 import NSGAII
from pyincore.analyses.transportationrecovery.post_disaster_long_term_solution import PostDisasterLongTermSolution
from pyincore.analyses.transportationrecovery import WIPW
from pyincore.analyses.transportationrecovery.transportationrecovery import TransportationRecovery
from pyincore.analyses.transportationrecovery.transportationrecoveryutil import TransportationRecoveryUtil
